/*

  FreeWRL support library.
  Texture handling code.

*/

/****************************************************************************
    This file is part of the FreeWRL/FreeX3D Distribution.

    Copyright 2009 CRC Canada. (http://www.crc.gc.ca)

    FreeWRL/FreeX3D is free software: you can redistribute it and/or modify
    it under the terms of the GNU Lesser Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    FreeWRL/FreeX3D is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with FreeWRL/FreeX3D.  If not, see <http://www.gnu.org/licenses/>.
****************************************************************************/



#include <config.h>
#include <system.h>
#include <system_threads.h>
#include <display.h>
#include <internal.h>

#include <libFreeWRL.h>
#include <list.h>

#include <threads.h>

#include "../vrml_parser/Structs.h"
#include "../main/headers.h"

#include "../scenegraph/readpng.h"
#include "../input/InputFunctions.h"
#include "../opengl/Material.h"
#include "../opengl/OpenGL_Utils.h"
#include "Textures.h"
#include "../world_script/fieldSet.h"
#include "../scenegraph/Component_Shape.h"
#include "../scenegraph/Component_CubeMapTexturing.h"
#include "../scenegraph/RenderFuncs.h"
#include "LoadTextures.h"

// OLD_IPHONE_AQUA #ifdef AQUA
// OLD_IPHONE_AQUA #ifndef IPHONE
// OLD_IPHONE_AQUA # include <Carbon/Carbon.h>
// OLD_IPHONE_AQUA # include <QuickTime/QuickTime.h>
// OLD_IPHONE_AQUA #endif
// OLD_IPHONE_AQUA #else
// OLD_IPHONE_AQUA # if HAVE_JPEGLIB_H
// OLD_IPHONE_AQUA #undef HAVE_STDLIB_H
// OLD_IPHONE_AQUA #undef FAR
// OLD_IPHONE_AQUA #  include <jpeglib.h>
// OLD_IPHONE_AQUA #  include <setjmp.h>
// OLD_IPHONE_AQUA # endif
// OLD_IPHONE_AQUA #endif



struct multiTexParams {
int multitex_mode[2];
int multitex_source[2];
int multitex_function;
};


#ifndef GL_EXT_texture_cube_map
//OLDCODE # define GL_NORMAL_MAP_EXT                   0x8511
//OLDCODE # define GL_REFLECTION_MAP_EXT               0x8512
# define GL_TEXTURE_CUBE_MAP_EXT             0x8513
//OLDCODE # define GL_TEXTURE_BINDING_CUBE_MAP_EXT     0x8514
# define GL_TEXTURE_CUBE_MAP_POSITIVE_X_EXT  0x8515
//OLDCODE # define GL_TEXTURE_CUBE_MAP_NEGATIVE_X_EXT  0x8516
//OLDCODE # define GL_TEXTURE_CUBE_MAP_POSITIVE_Y_EXT  0x8517
//OLDCODE # define GL_TEXTURE_CUBE_MAP_NEGATIVE_Y_EXT  0x8518
//OLDCODE # define GL_TEXTURE_CUBE_MAP_POSITIVE_Z_EXT  0x8519
//OLDCODE # define GL_TEXTURE_CUBE_MAP_NEGATIVE_Z_EXT  0x851A
//OLDCODE # define GL_PROXY_TEXTURE_CUBE_MAP_EXT       0x851B
//OLDCODE # define GL_MAX_CUBE_MAP_TEXTURE_SIZE_EXT    0x851C
#endif


static void new_bind_image(struct X3D_Node *node, struct multiTexParams *vparam);
textureTableIndexStruct_s *getTableIndex(int i);

typedef struct pTextures{
	struct Vector *activeTextureTable;
	textureTableIndexStruct_s* loadThisTexture;

	/* current index into loadparams that texture thread is working on */
	int currentlyWorkingOn;// = -1;
	int textureInProcess;// = -1;
}* ppTextures;




void *Textures_constructor(){
	void *v = MALLOCV(sizeof(struct pTextures));
	memset(v,0,sizeof(struct pTextures));
	return v;
}
void Textures_init(struct tTextures *t){
	//public

	//private
	t->prv = Textures_constructor();
	{
		ppTextures p = (ppTextures)t->prv;
		p->activeTextureTable = NULL;

		/* current index into loadparams that texture thread is working on */
		p->currentlyWorkingOn = -1;

		p->textureInProcess = -1;
	}
}
void Textures_clear(struct tTextures *t){
	//public
	glDeleteBuffers (1,&t->defaultBlankTexture);
	//private
	{
		//int i;
		ppTextures p = (ppTextures)t->prv;
		//for(i=0;i<vectorSize(p->activeTextureTable);i++){
		//	textureTableIndexStruct_s *tti = vector_get(textureTableIndexStruct_s *,p->activeTextureTable,i);
		//	if(tti && tti->texdata)
		//		FREE_IF_NZ(tti->texdata);
		//}
		deleteVector(textureTableIndexStruct_s *, p->activeTextureTable);
	}
}


// OLD_IPHONE_AQUA #if defined(AQUA) /* for AQUA OS X sharing of OpenGL Contexts */

// OLD_IPHONE_AQUA #elif defined(_MSC_VER)
#if defined(_MSC_VER) || defined(AQUA)


#else
#if !defined(_ANDROID) && !defined(ANDROIDNDK) && !defined(GLES2)
GLXContext textureContext = NULL;
#endif

#endif

/* function Prototypes */
int findTextureFile(textureTableIndexStruct_s *entry);
//void _textureThread(void);

void move_texture_to_opengl(textureTableIndexStruct_s*);
struct Uni_String *newASCIIString(const char *str);

int readpng_init(FILE *infile, ulg *pWidth, ulg *pHeight);
void readpng_cleanup(int free_image_data);


static void myTexImage2D (int generateMipMaps, GLenum target, GLint level, GLint internalformat, GLsizei width, GLsizei height, GLint border, GLenum format, GLenum type, GLubyte *pixels);

/* gluScaleImage replacement */

/* we manipulate things here as 4 byte colours (RGBA, or BGRA) so we try and ensure
that we have 32 bit numbers, thus this definition. ES-2.0 is 32 bits, but in case
someone compiles on a 64 bit system, with Material shader definitions in OpenGL-3.x,
then hopefully this will work for them, too */

#ifndef uint32
# define uint32 uint32_t
#endif



static void myScaleImage(int srcX,int srcY,int destX,int destY,unsigned char *src, unsigned char *dest) {
	float YscaleFactor;
	float XscaleFactor;
	int wye, yex;
	uint32 *src32 = (uint32 *)src;
	uint32 *dest32 = (uint32 *)dest;

	if ((srcY<=0) || (destY<=0) || (srcX<=0) || (destX<=0)) return;
	if (src == NULL) return;
	if (dest == NULL) return;

	if ((srcY==destY) && (srcX==destX)) {
		/* printf ("simple copy\n"); */
		memcpy (dest,src,srcY*srcX*4); /* assuming FreeWRL-standard RGBA or BGRA textures */
	}

	/* do x direction first */
	YscaleFactor = ((float)srcY) / ((float)destY);
	XscaleFactor = ((float)srcX) / ((float)destX);

	for (wye=0; wye<destY; wye++) {
		for (yex=0; yex<destX; yex++) {
			float fx, fy;
			int row, column;
			size_t oldIndex, newIndex;

			fx = YscaleFactor * ((float) wye);
			fy = XscaleFactor * ((float) yex);
			row = (int)(fx);
			column = (int)(fy);
			oldIndex = (size_t)row * (size_t)srcX + (size_t)column; /* so many rows, each row has srcX columns */
			newIndex = (size_t)wye * (size_t)destX + (size_t)yex; //wye*destX+yex
			dest32[newIndex] = src32[oldIndex];
		}
	}
}

static void myScaleImage3D(int srcX,int srcY,int srcZ, int destX,int destY,int destZ, unsigned char *src, unsigned char *dest) {
	float YscaleFactor;
	float XscaleFactor;
	float ZscaleFactor;
	float fx,fy,fz;
	int iy, ix, iz;
	uint32 *src32 = (uint32 *)src;
	uint32 *dest32 = (uint32 *)dest;

	if ((srcY<=0) || (destY<=0) || (srcX<=0) || (destX<=0) || (srcZ<=0) || (destZ<=0)) return;
	if (src == NULL) return;
	if (dest == NULL) return;

	if ((srcY==destY) && (srcX==destX) && (srcZ==destZ)) {
		/* printf ("simple copy\n"); */
		memcpy (dest,src,srcY*srcX*srcZ*4); /* assuming FreeWRL-standard RGBA or BGRA textures */
	}

	/* do x direction first */
	YscaleFactor = ((float)srcY) / ((float)destY);
	XscaleFactor = ((float)srcX) / ((float)destX);
	ZscaleFactor = ((float)srcZ) / ((float)destZ);

	for (iz=0; iz<destZ; iz++) {
		int page;
		fz = ZscaleFactor * ((float) iz);
		page = (int)fz;
		for (iy=0; iy<destY; iy++) {
			int row;
			fy = YscaleFactor * ((float) iy);
			row = (int)(fy);
			for (ix=0; ix<destX; ix++) {
				int column;
				int oldIndex;

				fx = XscaleFactor * ((float) ix);
				column = (int)(fx);
				oldIndex = (page * srcY + row) * srcX + column; /* so many rows, each row has srcX columns */
				dest32[(iz*destY + iy)*destX+ix] = src32[oldIndex];
			}
		}
	}
}
int iclamp(int ival, int istart, int iend) {
	int iret = ival;
	iret = ival > iend? iend : ival;
	iret = iret < istart ? istart : iret;
	return iret;
}
void compute_3D_alpha_gradient_store_rgb(char *dest,int x,int y, int z){
	//assumes we have a scalar image with info only in alpha, but (unused) RGB channels
	//we compute 3D alpha/scalar gradient using one of sobel, roberts ...
	//gradient has magnitude and direction (vs normal, which is of unit length)
	//here we do an axis-aligned roberts ie gradient_x = x1 - x0
	int iz,iy,ix, jz,jy,jx,jzz,jyy,jxx, k;
	char *rgba0, *rgba1;
	int gradient[3], maxgradient[3], mingradient[3];
	unsigned char *urgba;
	uint32 *pixels = (uint32 *)dest;

	for(k=0;k<3;k++) {
		maxgradient[k] = -1;
		mingradient[k] = 1;
	}


	for(iz=0;iz<z;iz++){
		for(iy=0;iy<y;iy++){
			for(ix=0;ix<x;ix++){
				//initialize gradient
				for(k=0;k<3;k++) gradient[k] = 0;
				rgba0 = (char *) &pixels[(iz*y +iy)*x + ix];
				urgba = (unsigned char *)rgba0;
				if(1){
					//sum onto gradient
					jxx = jyy = jzz = 0;
					//what if we are on the edge? for roberts, just duplicate next-to-edge by backing up one
					if(iz == z-1) jzz = -1;
					if(iy == y-1) jyy = -1;
					if(ix == x-1) jxx = -1;
					jx = jxx; jy = jyy; jz = jzz;
					jx = jxx + 1;
					rgba1 = (char *) &pixels[((iz+jz)*y +(iy+jy))*x + (ix+jx)];
					gradient[0] = (int)rgba1[3] - (int)rgba0[3];
					jx = jxx;
					jy = jyy+1;
					rgba1 = (char *) &pixels[((iz+jz)*y +(iy+jy))*x + (ix+jx)];
					gradient[1] = (int)rgba1[3] - (int)rgba0[3];
					jy = jyy;
					jz = jzz+1;
					rgba1 = (char *) &pixels[((iz+jz)*y +(iy+jy))*x + (ix+jx)];
					gradient[2] = (int)rgba1[3] - (int)rgba0[3];
				}else {
					//extract edge-clamped 3x3x3
					//       X  Y  Z with [1] in center
					int cube[3][3][3], i,j,ii,jj,kk;
					for(i=-1;i<2;i++){
						ii = iclamp(ix+i,0,x-1);
						for(j=-1;j<2;j++){
							jj= iclamp(iy+j,0,y-1);
							for(k=-1;k<2;k++){
								kk = iclamp(iz+k,0,z-1);
								cube[i+1][j+1][k+1] = ((unsigned char *)&pixels[(kk*y +jj)*x + ii])[3];
							}
						}
					}
					if(1){
						//roberts cross
						//gradient[0] = (cube[2][2][2] - cube[1][1][1]) + (cube[2][0][0] - cube[1][1][1];
					}
					if(1){
						//sobel gradient
						gradient[0] = 0;
						gradient[0] += cube[0][0][1] + 2*cube[0][1][1] + cube[0][2][1];
						gradient[0] -= cube[2][0][1] + 2*cube[2][1][1] + cube[2][2][1];
						gradient[0] += cube[0][1][0] + 2*cube[0][1][1] + cube[0][1][2];
						gradient[0] -= cube[2][1][0] + 2*cube[2][1][1] + cube[2][1][2];

						gradient[1] = 0;
						gradient[1] += cube[1][0][0] + 2*cube[1][0][1] + cube[1][0][2];
						gradient[1] -= cube[1][2][0] + 2*cube[1][2][1] + cube[1][2][2];
						gradient[1] += cube[0][0][1] + 2*cube[1][0][1] + cube[2][0][1];
						gradient[1] -= cube[9][2][1] + 2*cube[1][2][1] + cube[2][2][1];

						gradient[2] = 0;
						gradient[2] += cube[0][1][0] + 2*cube[1][1][0] + cube[2][1][0];
						gradient[2] -= cube[0][1][2] + 2*cube[1][1][2] + cube[2][1][2];
						gradient[2] += cube[1][0][0] + 2*cube[1][1][0] + cube[1][2][0];
						gradient[2] -= cube[1][9][2] + 2*cube[1][1][2] + cube[1][2][2];
						for(k=0;k<3;k++)
							gradient[k] /= 2;

					}
				}

				//scale gradient to -127 to +127 in each dimension
				//roberts: a1 - a0 could be in range (255 - 0) to (0 -255) or -255 to 255,
				// we need -127 to 127 signed char on each dim
				for(k=0;k<3;k++) gradient[k] /= 2;
				for(k=0;k<3;k++) {
					maxgradient[k] = max(maxgradient[k],gradient[k]);
					mingradient[k] = min(mingradient[k],gradient[k]);
				}

				//but when texture2D / sampler2D convert from image pixel to float,
				//the expect the pixels to be unsigned char.
				//so we add 127 here, and subtract .5 in the shader, once they are float
				for(k=0;k<3;k++) gradient[k] += 127;

				//set gradient in RGB channels
				for(k=0;k<3;k++) urgba[k] = (unsigned char)gradient[k];
				//if(rgba0[0] || rgba0[1] || rgba0[2]){
				//	if(rgba0[0] != rgba0[1] || rgba0[1] != rgba0[2])
				//		printf("[%d %d %d]",(int)rgba0[0],(int)rgba0[1],(int)rgba0[2]);
				//}
			}
		}
	}
	if(0){
	printf("mingradient %d %d %d\n",mingradient[0],mingradient[1],mingradient[2]);
	printf("maxgradient %d %d %d\n",maxgradient[0],maxgradient[1],maxgradient[2]);
	}
	if(0){
		//save gradient image for testing
		textureTableIndexStruct_s *tti2, tt;
		tti2 = &tt;
		tti2->x = x;
		tti2->y = y;
		tti2->z = z;
		tti2->texdata = (unsigned char *)dest;
		tti2->channels = 3;
		saveImage_web3dit(tti2,"gradientRGB.web3dit");
		tti2->channels = 4;
		saveImage_web3dit(tti2,"gradientRGBA.web3dit");

	}

}
static void GenMipMap2D( GLubyte *src, GLubyte **dst, int srcWidth, int srcHeight, int *dstWidth, int *dstHeight )
{
   int x,
       y;
   int texelSize = 4;
   size_t total_size;
   GLubyte *dest, *pix;


   *dstWidth = srcWidth / 2;
   if ( *dstWidth <= 0 )
      *dstWidth = 1;

   *dstHeight = srcHeight / 2;
   if ( *dstHeight <= 0 )
      *dstHeight = 1;

   total_size = sizeof(GLubyte) * (size_t)texelSize * (size_t)(*dstWidth) * (size_t)(*dstHeight);
   *dst = MALLOC(void *, total_size );
   dest = *dst;
   if ( *dst == NULL )
      return;

   for ( y = 0; y < *dstHeight; y++ )
   {
      for( x = 0; x < *dstWidth; x++ )
      {
         size_t srcIndex[4], x2, y2, swidth, texsize, kd;
         float r = 0.0f,
               g = 0.0f,
               b = 0.0f,
	       a = 0.0f;

         int sample;

        // Compute the offsets for 2x2 grid of pixels in previous
         // image to perform box filter
         //srcIndex[0] =
         //   (((y * 2) * srcWidth) + (x * 2)) * texelSize;
         //srcIndex[1] =
         //   (((y * 2) * srcWidth) + (x * 2 + 1)) * texelSize;
         //srcIndex[2] =
         //   ((((y * 2) + 1) * srcWidth) + (x * 2)) * texelSize;
         //srcIndex[3] =
         //   ((((y * 2) + 1) * srcWidth) + (x * 2 + 1)) * texelSize;

		 swidth = srcWidth;
		 texsize = texelSize;
		 y2 = y * 2L;
		 x2 = x * 2L;
         srcIndex[0] = (y2*swidth + x2) * texsize;
         srcIndex[1] = (y2*swidth + x2 + (size_t)1L) * texsize;
         srcIndex[2] = ((y2 + 1L)*swidth + x2) * texsize;
         srcIndex[3] = ((y2 + 1L)*swidth + x2 + (size_t)1L) * texsize;

         // Sum all pixels
         for ( sample = 0; sample < 4; sample++ )
         {
            r += src[srcIndex[sample] + (size_t)0L];
            g += src[srcIndex[sample] + (size_t)1L];
            b += src[srcIndex[sample] + (size_t)2L];
            a += src[srcIndex[sample] + (size_t)3L];
         }

         // Average results
         r /= 4.0f;
         g /= 4.0f;
         b /= 4.0f;
         a /= 4.0f;

         // Store resulting pixels
		 kd = ((size_t)y * (size_t)(*dstWidth) + (size_t)x ) * (size_t)texelSize;
		 pix = dest + kd;
         pix[0] = (GLubyte)( r );
         pix[1] = (GLubyte)( g );
         pix[2] = (GLubyte)( b );
         pix[3] = (GLubyte)( a );

         //(*dst)[ ( y * (*dstWidth) + x ) * texelSize ] = (GLubyte)( r );
         //(*dst)[ ( y * (*dstWidth) + x ) * texelSize + 1] = (GLubyte)( g );
         //(*dst)[ ( y * (*dstWidth) + x ) * texelSize + 2] = (GLubyte)( b );
         //(*dst)[ ( y * (*dstWidth) + x ) * texelSize + 3] = (GLubyte)( a );
      }
   }
}
/* create the MIPMAPS ourselves, as the OpenGL ES 2.0 can not do it */
static void myTexImage2D (int generateMipMaps, GLenum target, GLint level, GLint internalformat, GLsizei width, GLsizei height, GLint border, GLenum format, GLenum type, GLubyte *pixels) {
	GLubyte *prevImage = NULL;
	GLubyte *newImage = NULL;
	size_t total_size;

	/* first, base image */

	FW_GL_TEXIMAGE2D(target,level,internalformat,width,height,border,format,type,pixels);
	if (!generateMipMaps) return;
	if ((width <=1) && (height <=1)) return;


	/* go and create a bunch of mipmaps */
	total_size = (size_t)4L * width * height;
	prevImage = MALLOC(GLubyte *, total_size);
	memcpy (prevImage, pixels, total_size);

	/* from the OpenGL-ES 2.0 book, page 189 */
	level = 1;

	while ((width > 1) && (height > 1)) {
		GLint newWidth, newHeight;

		GenMipMap2D(prevImage,&newImage, width, height, &newWidth, &newHeight);

		FW_GL_TEXIMAGE2D(target,level,internalformat,newWidth,newHeight,0,format,GL_UNSIGNED_BYTE,newImage);

		FREE_IF_NZ(prevImage);
		prevImage = newImage;
		level++;
		width = newWidth;
		height = newHeight;
	}

	FREE_IF_NZ(newImage);

}


/**
 *   texst: texture status string.
 */
const char *texst(int num)
{
	if (num == TEX_NOTLOADED) return "TEX_NOTLOADED";
	if (num == TEX_LOADING) return "TEX_LOADING";
	if (num == TEX_NEEDSBINDING)return "TEX_NEEDSBINDING";
	if (num == TEX_LOADED)return "TEX_LOADED";
	if (num == TEX_UNSQUASHED)return "UNSQUASHED";
	return "unknown";
}


/* does a texture have alpha?  - pass in a __tableIndex from a MovieTexture, ImageTexture or PixelTexture. */
int isTextureAlpha(int texno) {
	textureTableIndexStruct_s *ti;

	/* no, have not even started looking at this */
	/* if (texno == 0) return FALSE; */

	ti = getTableIndex(texno);
	if (ti==NULL) return FALSE;

	if (ti->status==TEX_LOADED) {
		return ti->hasAlpha;
	}
	return FALSE;
}


/* is the texture thread initialized yet? */
int fwl_isTextureinitialized() {
	return gglobal()->threads.TextureThreadRunning; // gglobal()->LoadTextures.TextureThreadInitialized;
}

/* is this texture loaded? used in LoadSensor */
int fwl_isTextureLoaded(int texno) {
	textureTableIndexStruct_s *ti;

	/* no, have not even started looking at this */
	/* if (texno == 0) return FALSE; */

	ti = getTableIndex(texno);
	return (ti->status==TEX_LOADED);
}

/* statusbar uses this to tell user that we are still loading */
int fwl_isTextureParsing() {
	ppTextures p = (ppTextures)gglobal()->Textures.prv;

	/* return currentlyWorkingOn>=0; */
#ifdef TEXVERBOSE
    if (p->textureInProcess > 0) {
	printf("call to fwl_isTextureParsing %d, returning %d\n",
	       p->textureInProcess,p->textureInProcess > 0);
    }
#endif
	return p->textureInProcess >0;
}

/* this node has changed - if there was a texture, destroy it */
void releaseTexture(struct X3D_Node *node) {

	int tableIndex;
	textureTableIndexStruct_s *ti;

		if (node->_nodeType == NODE_ImageTexture) {
			tableIndex  = ((struct X3D_ImageTexture *)node)->__textureTableIndex;
		} else if (node->_nodeType == NODE_PixelTexture) {
			tableIndex  = ((struct X3D_PixelTexture *)node)->__textureTableIndex;
		} else if (node->_nodeType == NODE_BufferTexture) {
			tableIndex = ((struct X3D_BufferTexture*)node)->__textureTableIndex;
		} else if (node->_nodeType == NODE_MovieTexture) {
			tableIndex  = ((struct X3D_MovieTexture *)node)->__textureTableIndex;
		} else if (node->_nodeType == NODE_PixelTexture3D) {
			tableIndex  = ((struct X3D_PixelTexture3D *)node)->__textureTableIndex;
		} else if (node->_nodeType == NODE_ImageTexture3D) {
			tableIndex  = ((struct X3D_ImageTexture3D *)node)->__textureTableIndex;
		} else if (node->_nodeType == NODE_ComposedTexture3D) {
			tableIndex  = ((struct X3D_ComposedTexture3D *)node)->__textureTableIndex;

		} else return;

#ifdef TEXVERBOSE
	printf ("releaseTexture, calling getTableIndex\n");
	ti = getTableIndex(tableIndex);
	printf ("releaseTexture, ti %p, ti->status %d\n",ti,ti->status);
	ti->status = TEX_NOTLOADED;

	if (ti->OpenGLTexture != TEXTURE_INVALID) {
		printf ("deleting %d textures, starting at %u\n",ti->frames, ti->OpenGLTexture);
		ti->OpenGLTexture = TEXTURE_INVALID;
/* 		FREE_IF_NZ(ti->OpenGLTexture); */
	}
#endif

	ti = getTableIndex(tableIndex);
	if(ti){
		ti->status = TEX_NOTLOADED;
		if (ti->OpenGLTexture != TEXTURE_INVALID) {
			FW_GL_DELETETEXTURES(1, &ti->OpenGLTexture);
			ti->OpenGLTexture = TEXTURE_INVALID;
	/* 		FREE_IF_NZ(ti->OpenGLTexture); */
		}
	}
}



textureTableIndexStruct_s *getTableIndex(int indx) {
	ppTextures p = (ppTextures)gglobal()->Textures.prv;

#ifdef VERBOSE
{char line[200];
sprintf (line,"getTableIndex, looking for %d",indx);
ConsoleMessage (line);}
#endif
	if (indx < 0) {
		ConsoleMessage ("getTableIndex, texture <0 requested");
		return NULL;
	}

	if (p->activeTextureTable ==NULL ) {
		ConsoleMessage ("NULL sizing errror in getTableIndex");
		return NULL;
	}


	if (indx >= vectorSize(p->activeTextureTable)) {
		ConsoleMessage ("sizing errror in getTableIndex");
		return NULL;
	}


#ifdef VERBOSE
{char line[200];
printf ("getTableIndex - valid request\n");
sprintf (line,"getTableIndex, for %d, size %d",indx, vectorSize(p->activeTextureTable));
ConsoleMessage (line);}
#endif

	return vector_get(textureTableIndexStruct_s *, p->activeTextureTable, indx);
}
int getTextureTableIndexFromFromTextureNode(struct X3D_Node *node){
	int thisTexture = -1;
	int thisTextureType = node->_nodeType;
	if (thisTextureType==NODE_ImageTexture){
		struct X3D_ImageTexture* it = (struct X3D_ImageTexture*) node;
		thisTexture = it->__textureTableIndex;
	} else if (thisTextureType==NODE_PixelTexture){
		struct X3D_PixelTexture* pt = (struct X3D_PixelTexture*) node;
		thisTexture = pt->__textureTableIndex;
	} else if (thisTextureType == NODE_BufferTexture) {
		struct X3D_BufferTexture* pt = (struct X3D_BufferTexture*)node;
		thisTexture = pt->__textureTableIndex;
	} else if (thisTextureType == NODE_GeneratedTexture) {
		struct X3D_GeneratedTexture* gt = (struct X3D_GeneratedTexture*)node;
		thisTexture = gt->__textureTableIndex;
	} else if (thisTextureType==NODE_MovieTexture){
		struct X3D_MovieTexture* mt = (struct X3D_MovieTexture*) node;
		thisTexture = mt->__textureTableIndex;
	}else if (thisTextureType == NODE_ComposedCubeMapTexture) {
		struct X3D_ImageCubeMapTexture* ict = (struct X3D_ImageCubeMapTexture*)node;
		thisTexture = ict->__textureTableIndex;
	} else if (thisTextureType==NODE_ImageCubeMapTexture){
		struct X3D_ImageCubeMapTexture* ict = (struct X3D_ImageCubeMapTexture*) node;
		thisTexture = ict->__textureTableIndex;
	} else if (thisTextureType==NODE_GeneratedCubeMapTexture){
		struct X3D_GeneratedCubeMapTexture* ict = (struct X3D_GeneratedCubeMapTexture*) node;
		thisTexture = ict->__textureTableIndex;
	} else if (thisTextureType==NODE_PixelTexture3D){
		struct X3D_PixelTexture3D* pt = (struct X3D_PixelTexture3D*) node;
		thisTexture = pt->__textureTableIndex;
	} else if (thisTextureType==NODE_ImageTexture3D){
		struct X3D_ImageTexture3D* pt = (struct X3D_ImageTexture3D*) node;
		thisTexture = pt->__textureTableIndex;
	} else if (thisTextureType==NODE_ComposedTexture3D){
		struct X3D_ComposedTexture3D* pt = (struct X3D_ComposedTexture3D*) node;
		thisTexture = pt->__textureTableIndex;
	} else if (thisTextureType==NODE_MultiTexture){
		struct X3D_MultiTexture* pt = (struct X3D_MultiTexture*) node;
		thisTexture = getTextureTableIndexFromFromTextureNode(X3D_NODE(pt->texture.p[0]));
	} else {
		ConsoleMessage ("Invalid type for texture, %s\n",stringNodeType(thisTextureType));
	}
	return thisTexture;
}
textureTableIndexStruct_s *getTableTableFromTextureNode(struct X3D_Node *textureNode){
	textureTableIndexStruct_s *ret = NULL;
	int index = getTextureTableIndexFromFromTextureNode(textureNode);
	if(index > -1)
		ret = getTableIndex(index);
	return ret;
}
int getGlTextureNumberFromTextureNode(struct X3D_Node *textureNode){
	textureTableIndexStruct_s *tts = getTableTableFromTextureNode(textureNode);
	if(tts == NULL) return 0;
	return tts->OpenGLTexture;
}

int getTextureSizeFromTextureNode(struct X3D_Node *textureNode, int *ixyz){
	int iret;
	textureTableIndexStruct_s *tts = getTableTableFromTextureNode(textureNode);
	ixyz[0] = ixyz[1] = ixyz[2] = 0;
	iret = 0;
	if(tts){
		ixyz[0] = tts->x;
		ixyz[1] = tts->y;
		ixyz[2] = tts->z;
		iret = 1;
	}
	return iret;
}


/* is this node a texture node? if so, lets keep track of its textures. */
/* worry about threads - do not make anything reallocable */
void registerTexture0(int iaction, struct X3D_Node *tmp) {
	//iaction =1 add, =0 remove
	struct X3D_ImageTexture *it;


	it = (struct X3D_ImageTexture *) tmp;
	/* printf ("registerTexture, found a %s\n",stringNodeType(it->_nodeType));  */

	if ((it->_nodeType == NODE_ImageTexture) ||
		(it->_nodeType == NODE_PixelTexture) ||
		(it->_nodeType == NODE_BufferTexture) ||
		(it->_nodeType == NODE_GeneratedTexture) ||
		(it->_nodeType == NODE_ComposedCubeMapTexture) ||
		(it->_nodeType == NODE_ImageCubeMapTexture) ||
		(it->_nodeType == NODE_GeneratedCubeMapTexture) ||
		(it->_nodeType == NODE_PixelTexture3D) ||
		(it->_nodeType == NODE_ImageTexture3D) ||
		(it->_nodeType == NODE_ComposedTexture3D) ||
		(it->_nodeType == NODE_MovieTexture)
        ) {
		ppTextures p = (ppTextures)gglobal()->Textures.prv;
		if(iaction){
			//ADD
			// for the index, stored in the X3D node.
			int textureNumber;
			// new texture table entry. Zero all data
			textureTableIndexStruct_s * newTexture = MALLOC (textureTableIndexStruct_s *,sizeof (textureTableIndexStruct_s));
			memset(newTexture,0,sizeof(textureTableIndexStruct_s));
			newTexture->z = 1; //just texturing3D is > 1


			if (p->activeTextureTable == NULL) {
				p->activeTextureTable =newVector(textureTableIndexStruct_s *, 16);
				vector_pushBack(textureTableIndexStruct_s *, p->activeTextureTable, newTexture);
			}

			// keep track of which texture this one is.
			textureNumber = vectorSize(p->activeTextureTable);

			//{char line[200]; sprintf (line,"registerTexture textureNumber %d",textureNumber); ConsoleMessage(line);}

			DEBUG_TEX("CREATING TEXTURE NODE: type %d\n", it->_nodeType);
			/* I need to know the texture "url" here... */

			switch (it->_nodeType) {
			/* save this index in the scene graph node */
			case NODE_ImageTexture:
				it->__textureTableIndex = textureNumber;
				break;
			case NODE_PixelTexture: {
				struct X3D_PixelTexture *pt;
				pt = (struct X3D_PixelTexture *) tmp;
				pt->__textureTableIndex = textureNumber;
				break; }
			case NODE_BufferTexture: {
				struct X3D_BufferTexture* pt;
				pt = (struct X3D_BufferTexture*)tmp;
				pt->__textureTableIndex = textureNumber;
				break; }
			case NODE_GeneratedTexture: {
				struct X3D_GeneratedTexture* pt;
				pt = (struct X3D_GeneratedTexture*)tmp;
				pt->__textureTableIndex = textureNumber;
				break; }
			case NODE_PixelTexture3D: {
				struct X3D_PixelTexture3D *pt;
				pt = (struct X3D_PixelTexture3D *) tmp;
				pt->__textureTableIndex = textureNumber;
				break; }
			case NODE_ImageTexture3D: {
				struct X3D_ImageTexture3D *pt;
				pt = (struct X3D_ImageTexture3D *) tmp;
				pt->__textureTableIndex = textureNumber;
				break; }
			case NODE_ComposedTexture3D: {
				struct X3D_ComposedTexture3D *pt;
				pt = (struct X3D_ComposedTexture3D *) tmp;
				pt->__textureTableIndex = textureNumber;
				break; }
			case NODE_MovieTexture: {
				struct X3D_MovieTexture *mt;
				mt = (struct X3D_MovieTexture *) tmp;
				mt->__textureTableIndex = textureNumber;
				break; }

			case NODE_ComposedCubeMapTexture: {
				struct X3D_ComposedCubeMapTexture* v1t;
				v1t = (struct X3D_ComposedCubeMapTexture*)tmp;
				v1t->__textureTableIndex = textureNumber;
				break;
			}

			case NODE_ImageCubeMapTexture: {
				struct X3D_ImageCubeMapTexture *v1t;
				v1t = (struct X3D_ImageCubeMapTexture *) tmp;
				v1t->__textureTableIndex = textureNumber;
				break;
			}

			case NODE_GeneratedCubeMapTexture: {
				struct X3D_GeneratedCubeMapTexture *v1t;
				v1t = (struct X3D_GeneratedCubeMapTexture *) tmp;
				v1t->__textureTableIndex = textureNumber;
				break;
			}

			}

			/* set the scenegraphNode here */
			newTexture->nodeType = it->_nodeType;
			newTexture->scenegraphNode = X3D_NODE(tmp);
			newTexture->textureNumber = textureNumber;
			// save this to our texture table
			vector_pushBack(textureTableIndexStruct_s *, p->activeTextureTable, newTexture);
		}else{
			//REMOVE
			//we're using int indexes so we can't compact the vector to remove the element
			//we need to flag it some how, and many functions use tti *getTableIndex(int num)
			//and check if the returned value is null before trying to use it.
			//we'll try using NULL as the signal its deleted.
			textureTableIndexStruct_s * tti = NULL;
			int *textureNumber = NULL;

			releaseTexture(tmp); //Mar 23, 2015 added, to zap from gl texture name list (and its texture storage)
								//otherwise for geoLOD and inline, the OS MEM usage keeps going up after unload/load cycle

			switch (it->_nodeType) {
			/* save this index in the scene graph node */
			case NODE_ImageTexture:
				textureNumber = &it->__textureTableIndex;
				break;
			case NODE_PixelTexture: {
				struct X3D_PixelTexture *pt;
				pt = (struct X3D_PixelTexture *) tmp;
				textureNumber = &pt->__textureTableIndex;
				break; }
			case NODE_BufferTexture: {
				struct X3D_BufferTexture* pt;
				pt = (struct X3D_BufferTexture*)tmp;
				textureNumber = &pt->__textureTableIndex;
				break; }
			case NODE_GeneratedTexture: {
				struct X3D_GeneratedTexture* pt;
				pt = (struct X3D_GeneratedTexture*)tmp;
				textureNumber = &pt->__textureTableIndex;
				break; }
			case NODE_PixelTexture3D: {
				struct X3D_PixelTexture3D *pt;
				pt = (struct X3D_PixelTexture3D *) tmp;
				textureNumber = &pt->__textureTableIndex;
				break; }
			case NODE_ImageTexture3D: {
				struct X3D_ImageTexture3D *pt;
				pt = (struct X3D_ImageTexture3D *) tmp;
				textureNumber = &pt->__textureTableIndex;
				break; }
			case NODE_ComposedTexture3D: {
				struct X3D_ComposedTexture3D *pt;
				pt = (struct X3D_ComposedTexture3D *) tmp;
				textureNumber = &pt->__textureTableIndex;
				break; }
			case NODE_MovieTexture: {
				struct X3D_MovieTexture *mt;
				mt = (struct X3D_MovieTexture *) tmp;
				textureNumber = &mt->__textureTableIndex;
				break; }

			case NODE_ComposedCubeMapTexture: {
				struct X3D_ComposedCubeMapTexture* v1t;
				v1t = (struct X3D_ComposedCubeMapTexture*)tmp;
				textureNumber = &v1t->__textureTableIndex;
				break; }

			case NODE_ImageCubeMapTexture: {
				struct X3D_ImageCubeMapTexture *v1t;
				v1t = (struct X3D_ImageCubeMapTexture *) tmp;
				textureNumber = &v1t->__textureTableIndex;
				break; }

			case NODE_GeneratedCubeMapTexture: {
				struct X3D_GeneratedCubeMapTexture *v1t;
				v1t = (struct X3D_GeneratedCubeMapTexture *) tmp;
				textureNumber = &v1t->__textureTableIndex;
				break; }
			}
			if(textureNumber){
				tti = getTableIndex(*textureNumber);
				if(tti){
					// (*textureNumber) = -1; //is there a better flag?
					vector_set(textureTableIndexStruct_s *,p->activeTextureTable,*textureNumber,NULL);
					//unregister/unbind/deallocate anything else that was registered/bound/allocated above

					//Problem: when unloading an inline (including geoLOD inlines) with images that haven't yet loaded,
					// load_inline > unload_broto > unregister_broto_instance > unRegisterX3DAnyNode > unRegisterTexture > registerTexture0 (here)
					// if we zap the tti, then when the resource thread finishes downloading the image and goes to paste
					// into wheretoplacedata* which is a tti* that's been zapped, we crash.
					//Solution:
					// quick fix: don't zap tti here, leave a memory leak TRIED, WORKED AS BANDAID
					// proper fix: redesign resource fetch so it has a way to check if tti still exists, IMPLEMENTED Dec5,2014
					//  for example, have it use the table index number instead of a pointer directly to *tti,
					//  and here leave a NULL in the table at that index so resource thread can check if its been zapped
					FREE_IF_NZ(tti->texdata);
					//->filename is not strduped, just a pointer to actual_file which is freed in resource
					FREE_IF_NZ(tti);
				}
			}
		}
	} else {
		//ConsoleMessage ("registerTexture, ignoring this node");
	}
}
void registerTexture(struct X3D_Node *tmp) {
	registerTexture0(1,tmp);
}
void unRegisterTexture(struct X3D_Node *tmp) {
	registerTexture0(0,tmp);
}


void add_node_to_broto_context(struct X3D_Proto *currentContext,struct X3D_Node *node);
/* do TextureBackground textures, if possible */
void reallyDraw();
void push_render_geom(int igeom);
void pop_render_geom();
void loadBackgroundTextures (struct X3D_Background *node) {
	struct X3D_ImageTexture *thistex;
	struct X3D_TextureProperties *thistp;
	struct Multi_String thisurl;
	int count;

	/* initialization */
	struct textureVertexInfo mtf = {boxtex,2,GL_FLOAT,0,NULL,NULL};
	thisurl.n = 0; thisurl.p = NULL;
	thistex = NULL;
	s_shader_capabilities_t* me;
	struct matpropstruct * mat = getAppearanceProperties();
	me = mat->currentShaderProperties;
	mat->fw_FrontMaterial.type = MAT_UNLIT;
	for (count=0; count<6; count++) {
		/* go through these, back, front, top, bottom, right left */
		switch (count) {
			case 0: {thistex = X3D_IMAGETEXTURE(node->__frontTexture);  thisurl = node->frontUrl; break;}
			case 1: {thistex = X3D_IMAGETEXTURE(node->__backTexture);   thisurl = node->backUrl; break;}
			case 2: {thistex = X3D_IMAGETEXTURE(node->__topTexture);    thisurl = node->topUrl; break;}
			case 3: {thistex = X3D_IMAGETEXTURE(node->__bottomTexture); thisurl = node->bottomUrl; break;}
			case 4: {thistex = X3D_IMAGETEXTURE(node->__rightTexture);  thisurl = node->rightUrl; break;}
			case 5: {thistex = X3D_IMAGETEXTURE(node->__leftTexture);   thisurl = node->leftUrl; break;}
		}
		if (thisurl.n != 0 ) {
			/* we might have to create a "shadow" node for the image texture */
			if (thistex == NULL) {
				int i;
				thistex = createNewX3DNode(NODE_ImageTexture);
				thistp = createNewX3DNode (NODE_TextureProperties);
				if(node->_executionContext){
					add_node_to_broto_context(X3D_PROTO(node->_executionContext),X3D_NODE(thistex));
					add_node_to_broto_context(X3D_PROTO(node->_executionContext),X3D_NODE(thistp));
				}

				/* set up TextureProperties, and link it in */

				/* we use the generic TextureProperties - especially the GenerateMipMaps flag... */
				thistp->generateMipMaps = GL_FALSE; /* default settings, put here to ensure that */
								/* future changes to the spec do no harm */
				thistp->textureCompression = newASCIIString("FASTEST");
				thistp->borderWidth = 0;
				thistex->textureProperties = X3D_NODE(thistp);
				ADD_PARENT(X3D_NODE(thistp), X3D_NODE(thistex));

#ifdef TEXVERBOSE
				printf ("bg, creating shadow texture node url.n = %d\n",thisurl.n);
#endif

				/* copy over the urls */
				thistex->url.p = MALLOC(struct Uni_String **, sizeof (struct Uni_String) * thisurl.n);
				for (i=0; i<thisurl.n; i++) {
					thistex->url.p[i] = newASCIIString (thisurl.p[i]->strptr);
				}
				thistex->url.n = thisurl.n;

				switch (count) {
					case 0: {node->__frontTexture = X3D_NODE(thistex);  break;}
					case 1: {node->__backTexture = X3D_NODE(thistex);   break;}
					case 2: {node->__topTexture = X3D_NODE(thistex);    break;}
					case 3: {node->__bottomTexture = X3D_NODE(thistex); break;}
					case 4: {node->__rightTexture = X3D_NODE(thistex);  break;}
					case 5: {node->__leftTexture = X3D_NODE(thistex);   break;}
				}
			}

			/* we have an image specified for this face */
			clear_textureUnit_used(); //appearance.texture material.textureXXX, PTMs.texture all need TEXTURE0+ XXX, where xxx starts from 0
			clear_material_samplers(); //PTM and material.textureXXX share frag shader sampler2D textureUnit[16] array
			clear_materialparameters_per_draw_counts(); //especially diffuse texture counts which both appearance and material share

			gglobal()->RenderFuncs.textureStackTop = 0;
			/* render the proper texture */
			push_render_geom(1);
			//POSSIBLE_PROTO_EXPANSION(struct X3D_Node*, node->texture, tmpN);
			//tg->RenderFuncs.texturenode = (void*)tmpN;
			gglobal()->RenderFuncs.texturenode = (void*)thistex; //textureTransform_start needs the node
			render_node(X3D_NODE(thistex));
			pop_render_geom();
		    //OLDCODE FW_GL_COLOR3D(1.0,1.0,1.0);
			textureTransform_start();
			//glUniform1i(me->TextureUnit[0], 0);
			setupShaderB();

        	textureCoord_send(&mtf);
        	FW_GL_VERTEX_POINTER(3,GL_FLOAT,0,BackgroundVert);
        	FW_GL_NORMAL_POINTER(GL_FLOAT,0,Backnorms);

        	sendArraysToGPU (GL_TRIANGLES, count*6, 6);
			reallyDraw();
			textureTransform_end();

		}
	}
}
/* do TextureBackground textures, if possible */
void loadTextureBackgroundTextures (struct X3D_TextureBackground *node) {
	struct X3D_Node *thistex = NULL;
	struct X3D_TextureProperties *thistp = NULL;
	int count;
	struct textureVertexInfo mtf = {boxtex,2,GL_FLOAT,0,NULL,NULL};
	struct matpropstruct* mat = getAppearanceProperties();
	mat->fw_FrontMaterial.type = MAT_UNLIT;

	for (count=0; count<6; count++) {
		/* go through these, back, front, top, bottom, right left */
		switch (count) {
			case 0: {POSSIBLE_PROTO_EXPANSION(struct X3D_Node *, node->frontTexture,thistex);  break;}
			case 1: {POSSIBLE_PROTO_EXPANSION(struct X3D_Node *, node->backTexture,thistex);   break;}
			case 2: {POSSIBLE_PROTO_EXPANSION(struct X3D_Node *, node->topTexture,thistex);    break;}
			case 3: {POSSIBLE_PROTO_EXPANSION(struct X3D_Node *, node->bottomTexture,thistex); break;}
			case 4: {POSSIBLE_PROTO_EXPANSION(struct X3D_Node *, node->rightTexture,thistex);  break;}
			case 5: {POSSIBLE_PROTO_EXPANSION(struct X3D_Node *, node->leftTexture,thistex);   break;}
		}
		if (thistex != 0) {
			/* we have an image specified for this face */
			/* the X3D spec says that a X3DTextureNode has to be one of... */
			if ((thistex->_nodeType == NODE_ImageTexture) ||
			    (thistex->_nodeType == NODE_PixelTexture) ||
				(thistex->_nodeType == NODE_BufferTexture) ||
			    (thistex->_nodeType == NODE_MovieTexture) ||
			    (thistex->_nodeType == NODE_MultiTexture)) {

				/* if we have no texture properties, add one here to disable mipmapping */
				switch (thistex->_nodeType) {
					case NODE_ImageTexture: {
						if (X3D_IMAGETEXTURE(thistex)->textureProperties == NULL) {
							thistp = createNewX3DNode (NODE_TextureProperties);
							if(node->_executionContext){
								add_node_to_broto_context(X3D_PROTO(node->_executionContext),X3D_NODE(thistp));
							}
							X3D_IMAGETEXTURE(thistex)->textureProperties = X3D_NODE(thistp);
							ADD_PARENT(X3D_NODE(thistp),thistex);
						}
						break;
					}
					case NODE_PixelTexture: {
						if (X3D_PIXELTEXTURE(thistex)->textureProperties == NULL) {
							thistp = createNewX3DNode (NODE_TextureProperties);
							if(node->_executionContext){
								add_node_to_broto_context(X3D_PROTO(node->_executionContext),X3D_NODE(thistp));
							}

							X3D_PIXELTEXTURE(thistex)->textureProperties = X3D_NODE(thistp);
							ADD_PARENT(X3D_NODE(thistp),thistex);
						}
						break;
					}
					case NODE_BufferTexture: {
						if (X3D_BUFFERTEXTURE(thistex)->textureProperties == NULL) {
							thistp = createNewX3DNode(NODE_TextureProperties);
							if (node->_executionContext) {
								add_node_to_broto_context(X3D_PROTO(node->_executionContext), X3D_NODE(thistp));
							}

							X3D_BUFFERTEXTURE(thistex)->textureProperties = X3D_NODE(thistp);
							ADD_PARENT(X3D_NODE(thistp), thistex);
						}
						break;
					}

					case NODE_MovieTexture:
					case NODE_MultiTexture:
					break;
				};


				clear_textureUnit_used(); //appearance.texture material.textureXXX, PTMs.texture all need TEXTURE0+ XXX, where xxx starts from 0
				clear_material_samplers(); //PTM and material.textureXXX share frag shader sampler2D textureUnit[16] array
				clear_materialparameters_per_draw_counts(); //especially diffuse texture counts which both appearance and material share

				gglobal()->RenderFuncs.textureStackTop = 0;
				/* render the proper texture */
				push_render_geom(1);
				//POSSIBLE_PROTO_EXPANSION(struct X3D_Node*, node->texture, tmpN);
				//tg->RenderFuncs.texturenode = (void*)tmpN;
				gglobal()->RenderFuncs.texturenode = (void*)thistex; //textureTransform_start needs the node
				render_node(X3D_NODE(thistex));
				pop_render_geom();
				//OLDCODE FW_GL_COLOR3D(1.0,1.0,1.0);
				textureTransform_start();
				//glUniform1i(me->TextureUnit[0], 0);
				setupShaderB();

				textureCoord_send(&mtf);
				FW_GL_VERTEX_POINTER(3, GL_FLOAT, 0, BackgroundVert);
				FW_GL_NORMAL_POINTER(GL_FLOAT, 0, Backnorms);

				sendArraysToGPU(GL_TRIANGLES, count * 6, 6);
				reallyDraw();
				textureTransform_end();

			}
		}
	}
}


/* load in a texture, if possible */
void loadTextureNode (struct X3D_Node *node, void *vparam)
{
	PRINT_GL_ERROR_IF_ANY("loadTextureNode")

    //printf ("loadTextureNode, node %p, params %p",node,param);
    if (NODE_NEEDS_COMPILING) {

		DEBUG_TEX ("FORCE NODE RELOAD: %p %s\n", node, stringNodeType(node->_nodeType));

		/* force a node reload - make it a new texture. Don't change
		   the parameters for the original number, because if this
		   texture is shared, then ALL references will change! so,
		   we just accept that the current texture parameters have to
		   be left behind. */
		MARK_NODE_COMPILED;

		/* this will cause bind_image to create a new "slot" for this texture */
		/* cast to GLuint because __texture defined in VRMLNodes.pm as SFInt */

		switch (node->_nodeType) {

			case NODE_MovieTexture: {
	    		//releaseTexture(node);
			}
			break;

			case NODE_PixelTexture:
	    			releaseTexture(node);
			break;
			case NODE_BufferTexture:
				releaseTexture(node);
				break;

			case NODE_ImageTexture:
	    			releaseTexture(node);
			break;
			case NODE_GeneratedTexture:
				releaseTexture(node);
				break;

			case NODE_ImageCubeMapTexture:
	    			releaseTexture(node);
			break;

			case NODE_GeneratedCubeMapTexture:
	    			//releaseTexture(node);
			break;

			case NODE_PixelTexture3D:
	    			releaseTexture(node);
			break;
			case NODE_ImageTexture3D:
	    			releaseTexture(node);
			break;
			case NODE_ComposedTexture3D:
	    			releaseTexture(node);
			break;

			default: {
				printf ("loadTextureNode, unknown node type %s\n",stringNodeType(node->_nodeType));
				return;
			}
	    }
	}

	new_bind_image (X3D_NODE(node), (struct multiTexParams *)vparam);

	return;
}

static void compileMultiTexture (struct X3D_MultiTexture *node) {
	struct multiTexParams *paramPtr;
	int count;
	int max;
	s_renderer_capabilities_t *rdr_caps;
	ttglobal tg = gglobal();
	rdr_caps = tg->display.rdr_caps;

	/*  have to regen the shape*/
	MARK_NODE_COMPILED;

	/* alloc fields, if required - only do this once, even if node changes */
	if (node->__xparams == 0) {
		/* printf ("loadMulti, MALLOCing for params\n"); */
		node->__xparams = MALLOC (void *, sizeof (struct multiTexParams) * rdr_caps->texture_units);

		 //printf ("just mallocd %ld in size for __params\n",sizeof (struct multiTexParams) * gglobal()->display.rdr_caps.texture_units);

		//printf ("paramPtr is %p\n",(int *)node->__params);

		paramPtr = (struct multiTexParams*) node->__xparams;

		/* set defaults for these fields */
		for (count = 0; count < rdr_caps->texture_units; count++) {
			paramPtr->multitex_mode[0]= MTMODE_MODULATE; //rgba (or rgb if a > 0)
			paramPtr->multitex_mode[1]= 0; //0=unused -1=default else alpha channel part
			paramPtr->multitex_source[0]=INT_ID_UNDEFINED; //rgba (or rgb if a > 0)
			paramPtr->multitex_source[1]=0; //0=unused -1=default else alpha channel part
			paramPtr->multitex_function=INT_ID_UNDEFINED;
			paramPtr++;
		}
	}

	/* how many textures can we use? no sense scanning those we cant use */
	//max = node->mode.n;
	max = node->texture.n;
	if (max > rdr_caps->texture_units) max = rdr_caps->texture_units;

	// warn users that function and source parameters not looked at right now
	//if ((node->source.n>0) || (node->function.n>0)) {
	//    ConsoleMessage ("currently, MultiTexture source and function parameters defaults used");
	//}
	/* go through the params, and change string name into an int */
	paramPtr = (struct multiTexParams*) node->__xparams;
	for (count = 0; count < max; count++) {
		char *smode, *ssource, *sfunc;
		smode = ssource = sfunc = NULL;
		if(node->mode.n>count){
			int mode, modea;
			smode = node->mode.p[count]->strptr;
			modea = 0; //unused
			mode = findFieldInMULTITEXTUREMODE(smode); //we offset by 1 in the #defines
			if(mode > -1) mode += 1; //one-based defines
			if(mode == -1){
				//might be Castle style "RGB / ALPHA" dual modes
				if(strchr(smode,'/') || strchr(smode,',')){
					//yes, castle protocol
					char *srgb, *salpha, *b1,*b2, *splittable;
					int modergb, modealpha;
					splittable = strdup(smode);
					b1 = strchr(splittable,' ');
					b2 = strrchr(splittable,' ');
					salpha = b2+1;
					splittable[b1 - splittable] = '\0';
					srgb = splittable;
					modergb = findFieldInMULTITEXTUREMODE(srgb);
					if(modergb == -1)
						modergb = MTMODE_MODULATE;
					else
						modergb +=1; //one-based defines
					modealpha = findFieldInMULTITEXTUREMODE(salpha);
					if(modealpha == -1)
						modealpha = MTMODE_MODULATE; //default
					else
						modealpha += 1; //one-based defines
					free(splittable);
					mode = modergb;
					modea = modealpha;
				}
			}
			if(mode > -1){
				paramPtr->multitex_mode[0] = mode;
				paramPtr->multitex_mode[1] = modea;
			}
			//else default
		}
		if(node->source.n > count) {
			int source, sourcea;
			ssource = node->source.p[count]->strptr;
			source = findFieldInMULTITEXTURESOURCE(ssource); //we offset by 1 in the #defines
			if(source > -1) source += 1; //one-based defines
			sourcea = 0; //0=unused
			if(source == -1){
				//might be Castle style "RGB / ALPHA" dual modes
				if(strchr(ssource,'/') || strchr(ssource,',')){
					//yes, castle protocol
					char *srgb, *salpha, *b1,*b2, *splittable;
					int sourcergb, sourcealpha;
					splittable = strdup(ssource);
					b1 = strchr(splittable,' ');
					b2 = strrchr(splittable,' ');
					salpha = b2+1;
					splittable[b1 - splittable] = '\0';
					srgb = splittable;
					sourcergb = findFieldInMULTITEXTURESOURCE(srgb);
					if(sourcergb == -1)
						sourcergb = INT_ID_UNDEFINED; //default
					else
						sourcergb +=1; //one-based defines
					sourcealpha = findFieldInMULTITEXTURESOURCE(salpha);
					free(splittable);
					source = sourcergb;
					sourcea = sourcealpha;
				}
			}
			if(source > -1){
				paramPtr->multitex_source[0] = source;
				paramPtr->multitex_source[1] = sourcea;
			}
			//else default
		}

		if (node->function.n>count) {
			int ifunc;
			sfunc = node->function.p[count]->strptr;
			{
			ifunc = findFieldInMULTITEXTUREFUNCTION(sfunc);
			}
			if(ifunc > -1)
				paramPtr->multitex_function = ifunc;
			//else default
		}

#ifdef TEXVERBOSE
printf ("compile_MultiTexture, %d of %d, mode %d %d source %d %d function %d m %s s %s f %s\n",
count,max,paramPtr->multitex_mode[0],paramPtr->multitex_mode[1],paramPtr->multitex_source[0],paramPtr->multitex_source[1],paramPtr->multitex_function,smode,ssource,sfunc);
#endif //TEXVERBOSE

		paramPtr++;
		}
	//printf("end of compileMultiTexture\n");
}

void loadMultiTexture (struct X3D_MultiTexture *node) {
	int count;
	int max;
	struct multiTexParams *paramPtr;
	struct X3D_ImageTexture *nt;
	s_renderer_capabilities_t *rdr_caps;
    ttglobal tg = gglobal();
	rdr_caps = tg->display.rdr_caps;

#ifdef TEXVERBOSE
	 printf ("loadMultiTexture, this %s has %d textures %x %x\n",stringNodeType(node->_nodeType),
			node->texture.n,
			(int) node->texture.p[0], (int) node->texture.p[1]);
	printf ("	change %d ichange %d\n",node->_change, node->_ichange);
#endif

	/* new node, or node paramaters changed */
        if (NODE_NEEDS_COMPILING) {
            compileMultiTexture(node);
	}

	/* ok, normally the scene graph contains function pointers. What we have
	   here is a set of pointers to datastructures of (hopefully!)
	   types like X3D_ImageTexture, X3D_PixelTexture, and X3D_MovieTexture.

	*/

	/* how many textures can we use? */
	max = node->texture.n;
    //printf ("texture.n %d, texture_units %d, MAX_MULTITEXTURE %d\n", node->texture.n, gglobal()->display.rdr_caps.texture_units, MAX_MULTITEXTURE);

	if (max > rdr_caps->texture_units) max = rdr_caps->texture_units;
    if (max > MAX_MULTITEXTURE) max = MAX_MULTITEXTURE;


	/* go through and get all of the textures */
	paramPtr = (struct multiTexParams *) node->__xparams;

#ifdef TEXVERBOSE
    printf ("loadMultiTExture, param stack:\n");
    for (count=0; count<max; count++) {
        printf ("   tex %d source %d mode %d\n",count,paramPtr[count].multitex_source,paramPtr[count].multitex_mode);
    }
#endif

	for (count=0; count < max; count++) {
#ifdef TEXVERBOSE
		printf ("loadMultiTexture, working on texture %d\n",count);
#endif

		/* get the texture */
		nt = X3D_IMAGETEXTURE(node->texture.p[count]);

		switch (nt->_nodeType) {
			case NODE_PixelTexture:
			case NODE_ImageTexture :
			case NODE_GeneratedTexture:
				/* printf ("MultiTexture %d is a ImageTexture param %d\n",count,*paramPtr);  */
				//loadTextureNode (X3D_NODE(nt),paramPtr);
				render_node(X3D_NODE(nt));
				break;
			case NODE_ImageCubeMapTexture:
			case NODE_ComposedCubeMapTexture:
			case NODE_GeneratedCubeMapTexture:
				//loadTextureNode(X3D_NODE(nt), paramPtr);
				render_node(X3D_NODE(nt));
				break;

			case NODE_MultiTexture:
				printf ("MultiTexture texture %d is a MULTITEXTURE!!\n",count);
				break;
			default:
				printf ("MultiTexture - unknown sub texture type %d\n",
						nt->_nodeType);
		}

		/* now, lets increment textureStackTop. The current texture will be
		   stored in boundTextureStack[textureStackTop]; textureStackTop will be 1
		   for "normal" textures; at least 1 for MultiTextures. */

        	tg->RenderFuncs.textureStackTop++;



        paramPtr++;

#ifdef TEXVERBOSE
        printf ("loadMultiTexture, textureStackTop %d\n",gglobal()->RenderFuncs.textureStackTop);
		printf ("loadMultiTexture, finished with texture %d\n",count);
#endif
	}
	tg->RenderFuncs.texturenode = (void*)node;
	//tg->RenderFuncs.multitexturenode = (void*)node;
}

int getTextureDescriptors(struct X3D_Node *textureNode, int *textures, int *modes, int *sources, int *funcs, int *width, int *height, int *samplr){
	int ntexture = 0;
	if( textureNode == NULL) return 0;
	if(textureNode->_nodeType == NODE_MultiTexture){
		struct multiTexParams *xparam;
		struct X3D_MultiTexture* pt = (struct X3D_MultiTexture*) textureNode;
		xparam = (struct multiTexParams *)pt->__xparams;
		if (!xparam) return 0;
		ntexture = pt->texture.n;
		for(int i=0;i<ntexture;i++){
			int iret, ixyz[3];

			textures[i] = getGlTextureNumberFromTextureNode(pt->texture.p[i]);
			iret = getTextureSizeFromTextureNode(pt->texture.p[i],ixyz);

			modes[i] = xparam[i].multitex_mode[0] + 100*xparam[i].multitex_mode[1];
			sources[i] = xparam[i].multitex_source[0] + 100*xparam[i].multitex_source[1];
			funcs[i] = xparam[i].multitex_function;
			samplr[i] = is_cubeMap(pt->texture.p[i]);
		}
	}else if (textureNode->_nodeType == NODE_ComposedTexture3D) {
		struct X3D_ComposedTexture3D* pt = (struct X3D_ComposedTexture3D*)textureNode;
		struct Multi_Node* tex = &pt->texture;
		ntexture = tex->n;
		for (int i = 0; i < ntexture; i++) {
			int iret, ixyz[3];

			textures[i] = getGlTextureNumberFromTextureNode(tex->p[i]);
			iret = getTextureSizeFromTextureNode(tex->p[i], ixyz);

			modes[i] = MTMODE_REPLACE;
			sources[i] = INT_ID_UNDEFINED;
			funcs[i] = INT_ID_UNDEFINED;
			samplr[i] = 0;
		}
	} else {
		//single texture, use web3d default texture descriptor
		int iret, ixyz[3];
		ntexture = 1;
		textures[0] = getGlTextureNumberFromTextureNode(textureNode);
		//web3d.org defaults? not sure, I think its replace
		modes[0] = MTMODE_REPLACE;
		sources[0] = INT_ID_UNDEFINED;
		funcs[0] = INT_ID_UNDEFINED;
		iret = getTextureSizeFromTextureNode(textureNode,ixyz);
		width[0] = ixyz[0];
		height[0] = ixyz[1];
		samplr[0] = is_cubeMap(textureNode);
	}
	return ntexture;
}


#define BOUNDARY_TO_GL(direct) \
				switch (findFieldInTEXTUREBOUNDARYKEYWORDS(tpNode->boundaryMode##direct->strptr)) { \
					case TB_CLAMP: direct##rc=GL_CLAMP; break; \
					case TB_CLAMP_TO_EDGE: direct##rc=GL_CLAMP_TO_EDGE; break; \
					case TB_CLAMP_TO_BOUNDARY: direct##rc=GL_CLAMP_TO_BORDER; break; \
					case TB_MIRRORED_REPEAT: direct##rc=GL_MIRRORED_REPEAT; break; \
					case TB_REPEAT: direct##rc=GL_REPEAT; break; \
					default: direct##rc = GL_REPEAT; \
				}

/* do we do 1 texture, or is this a series of textures, requiring final binding
   by this thread? */
#define DEF_FINDFIELD(arr) \
 static int findFieldIn##arr(const char* field) \
 { \
  return findFieldInARR(field, arr, arr##_COUNT); \
 }

DEF_FINDFIELD(TEXTUREMINIFICATIONKEYWORDS)
DEF_FINDFIELD(TEXTUREMAGNIFICATIONKEYWORDS)
DEF_FINDFIELD(TEXTUREBOUNDARYKEYWORDS)
DEF_FINDFIELD(TEXTURECOMPRESSIONKEYWORDS)

void unpackImageCubeMap6 (textureTableIndexStruct_s* me);
void move_texture_to_opengl(textureTableIndexStruct_s* me) {
	int rx, ry, rz, sx, sy, sz;
	int x, y, z;
	GLint iformat;
	GLenum format;

	/* default texture properties; can be changed by a TextureProperties node */
	float anisotropicDegree = 1.0f;
	int borderWidth;

	GLint Trc, Src, Rrc;
	GLint minFilter, magFilter;
	GLint compression;
	int generateMipMaps;

	unsigned char* mytexdata;

	// JAS - if multi-threading using a creation program,
	// we can have issues sending textures here, so
	// single thread this.
#ifdef PATH_PLANNER
	pthread_mutex_t gl_mutex1 = PTHREAD_MUTEX_INITIALIZER;
#endif //PATH_PLANNER

	/* for getting repeatS and repeatT info. */
	struct X3D_PixelTexture* pt = NULL;
	struct X3D_BufferTexture* bt = NULL;
	struct X3D_MovieTexture* mt = NULL;
	struct X3D_ImageTexture* it = NULL;
	struct X3D_PixelTexture3D* pt3d = NULL;
	struct X3D_GeneratedTexture* gt = NULL;
	struct X3D_TextureProperties* tpNode = NULL;
	int haveValidTexturePropertiesNode;
	GLfloat texPri;
	struct SFColorRGBA borderColour;
	s_renderer_capabilities_t* rdr_caps;

#ifdef PATH_PLANNER
	pthread_mutex_lock(&gl_mutex1);
#endif //PATH_PLANNER

	ttglobal tg = gglobal();
	rdr_caps = tg->display.rdr_caps;


	/* initialization */
	Src = FALSE; Trc = FALSE; Rrc = FALSE;
	tpNode = NULL;
	haveValidTexturePropertiesNode = FALSE;
	texPri = 0.0f;
	borderColour.c[0] = 0.0f; borderColour.c[1] = 0.0f; borderColour.c[2] = 0.0f; borderColour.c[3] = 0.0f;
	compression = GL_FALSE;
	borderWidth = 0;
	mytexdata = NULL;

	/* did this node get killed on the way here? */
	if (!checkNode(me->scenegraphNode, __FILE__, __LINE__)) {
		ConsoleMessage("main node disappeared, ignoring texture\n");
		me->status = TEXTURE_INVALID;

#ifdef PATH_PLANNER
		pthread_mutex_unlock(&gl_mutex1);
#endif //PATH_PLANNER

		return;
	}
	/* printf ("move_texture_to_opengl, node of type %s\n",stringNodeType(me->scenegraphNode->_nodeType));  */

	/* is this texture invalid and NOT caught before here? */
	/* this is the same as the defaultBlankTexture; the following code should NOT be executed */
	if (me->texdata == NULL) {
		char buff[] = { 0x70, 0x70, 0x70, 0xff }; /* same format as ImageTextures - GL_BGRA or GL_RGBA here */
		me->x = 1;
		me->y = 1;
		me->z = 1;
		me->hasAlpha = FALSE;
		me->texdata = MALLOC(unsigned char*, 4);
		memcpy(me->texdata, buff, 4);
	}

	/* do we need to convert this to an OpenGL texture stream?*/

//printf ("JAS - moveTex, texture %d, TEXTURE_INVALID %d at %s:%d\n",me->OpenGLTexture,TEXTURE_INVALID,__FILE__,__LINE__);

	/* we need to get parameters. */
	if (me->OpenGLTexture == TEXTURE_INVALID) {
		/* 		me->OpenGLTexture = MALLOC (GLuint *, sizeof (GLuint) * me->frames); */
		if ((getAppearanceProperties()->cubeFace == 0) || (getAppearanceProperties()->cubeFace == GL_TEXTURE_CUBE_MAP_POSITIVE_X)) {
			FW_GL_GENTEXTURES(1, &me->OpenGLTexture);
		}

#ifdef TEXVERBOSE
		printf("just glGend texture for block %p is %u, type %s\n",
			me, me->OpenGLTexture, stringNodeType(me->nodeType));
#endif

	}

	/* get the repeatS and repeatT info from the scenegraph node */
	if (me->nodeType == NODE_ImageTexture) {
		it = (struct X3D_ImageTexture*)me->scenegraphNode;
		Src = it->repeatS; Trc = it->repeatT;
		tpNode = X3D_TEXTUREPROPERTIES(it->textureProperties);
	}
	else if (me->nodeType == NODE_PixelTexture) {
		pt = (struct X3D_PixelTexture*)me->scenegraphNode;
		Src = pt->repeatS; Trc = pt->repeatT;
		tpNode = X3D_TEXTUREPROPERTIES(pt->textureProperties);
	}
	else if (me->nodeType == NODE_BufferTexture) {
		bt = (struct X3D_BufferTexture*)me->scenegraphNode;
		Src = bt->repeatS; Trc = bt->repeatT;
		tpNode = X3D_TEXTUREPROPERTIES(bt->textureProperties);
	}
	else if (me->nodeType == NODE_GeneratedTexture) {
		gt = (struct X3D_GeneratedTexture*)me->scenegraphNode;
		Src = gt->repeatS; Trc = gt->repeatT;
		tpNode = X3D_TEXTUREPROPERTIES(gt->textureProperties);
	}
	else if (me->nodeType == NODE_MovieTexture) {
		mt = (struct X3D_MovieTexture*)me->scenegraphNode;
		Src = mt->repeatS; Trc = mt->repeatT;
		tpNode = X3D_TEXTUREPROPERTIES(mt->textureProperties);
	}
	else if (me->nodeType == NODE_PixelTexture3D) {
		pt3d = (struct X3D_PixelTexture3D*)me->scenegraphNode;
		Src = pt3d->repeatS; Trc = pt3d->repeatT; Rrc = pt3d->repeatR;
		tpNode = X3D_TEXTUREPROPERTIES(pt3d->textureProperties);
	}
	else if (me->nodeType == NODE_ImageTexture3D) {
		struct X3D_ImageTexture3D* it3d;
		it3d = (struct X3D_ImageTexture3D*)me->scenegraphNode;
		Src = it3d->repeatS; Trc = it3d->repeatT; Rrc = it3d->repeatR;
		tpNode = X3D_TEXTUREPROPERTIES(it3d->textureProperties);
	}
	else if (me->nodeType == NODE_ComposedTexture3D) {
		struct X3D_ComposedTexture3D* ct3d;
		ct3d = (struct X3D_ComposedTexture3D*)me->scenegraphNode;
		Src = ct3d->repeatS; Trc = ct3d->repeatT; Rrc = ct3d->repeatR;
		tpNode = X3D_TEXTUREPROPERTIES(ct3d->textureProperties);
	}
	else if (me->nodeType == NODE_ImageCubeMapTexture) {
		struct X3D_ImageCubeMapTexture* mi = (struct X3D_ImageCubeMapTexture*)me->scenegraphNode;
		tpNode = X3D_TEXTUREPROPERTIES(mi->textureProperties);
	}
	else if (me->nodeType == NODE_GeneratedCubeMapTexture) {
		struct X3D_GeneratedCubeMapTexture* mi = (struct X3D_GeneratedCubeMapTexture*)me->scenegraphNode;
		tpNode = X3D_TEXTUREPROPERTIES(mi->textureProperties);
	}
	//texure3D faked via texture2D (in non-extended GLES2)
	//.. needs repeats for manual wrap vs clamp, will send in
	//.. passedInGenTex
	me->repeatSTR[0] = Src;
	me->repeatSTR[1] = Trc;
	me->repeatSTR[2] = Rrc;


	/* do we have a TextureProperties node? */
	if (tpNode) {
		if (tpNode->_nodeType != NODE_TextureProperties) {
			ConsoleMessage("have a %s as a textureProperties node", stringNodeType(tpNode->_nodeType));
		}
		else {
			haveValidTexturePropertiesNode = TRUE;
			generateMipMaps = tpNode->generateMipMaps ? GL_TRUE : GL_FALSE;
			texPri = tpNode->texturePriority;
			if ((texPri < 0.0) || (texPri > 1.0)) {
				texPri = 0.0f;
				ConsoleMessage("invalid texturePriority of %f", tpNode->texturePriority);
			}
			memcpy(&borderColour, &(tpNode->borderColor), sizeof(struct SFColorRGBA));

			anisotropicDegree = tpNode->anisotropicDegree;
			if ((anisotropicDegree < 1.0) || (anisotropicDegree > rdr_caps->anisotropicDegree)) {
				/* we can be quiet here
				   ConsoleMessage ("anisotropicDegree error %f, must be between 1.0 and %f",anisotropicDegree, gglobal()->display.rdr_caps.anisotropicDegree);
				*/
				anisotropicDegree = rdr_caps->anisotropicDegree;
			}

			borderWidth = tpNode->borderWidth;
			if (borderWidth < 0) borderWidth = 0;
			if (borderWidth > 1) borderWidth = 1;

			// http://www.web3d.org/documents/specifications/19775-1/V3.3/Part01/components/texturing.html#t-TextureMagnificationModes

			switch (findFieldInTEXTUREMAGNIFICATIONKEYWORDS(tpNode->magnificationFilter->strptr)) {
			case TMAG_AVG_PIXEL:
				magFilter = GL_LINEAR; break; // GL_NEAREST; break;
			case TMAG_DEFAULT: magFilter = GL_LINEAR; break;
			case TMAG_FASTEST: magFilter = GL_LINEAR; break;  /* DEFAULT */
			case TMAG_NEAREST_PIXEL: magFilter = GL_NEAREST; break;
			case TMAG_NICEST: magFilter = GL_NEAREST; break;
			default: magFilter = GL_NEAREST; ConsoleMessage("unknown magnification filter %s",
				tpNode->magnificationFilter->strptr);
			}

			/* minFilter depends on Mipmapping */
			if (generateMipMaps) switch (findFieldInTEXTUREMINIFICATIONKEYWORDS(tpNode->minificationFilter->strptr)) {
			case TMIN_AVG_PIXEL: minFilter = GL_NEAREST; break;
			case TMIN_AVG_PIXEL_AVG_MIPMAP: minFilter = GL_NEAREST_MIPMAP_NEAREST; break;
			case TMIN_AVG_PIXEL_NEAREST_MIPMAP: minFilter = GL_NEAREST_MIPMAP_NEAREST; break;
			case TMIN_DEFAULT: minFilter = GL_NEAREST_MIPMAP_LINEAR; break;
			case TMIN_FASTEST: minFilter = GL_NEAREST_MIPMAP_LINEAR; break;
			case TMIN_NICEST: minFilter = GL_NEAREST_MIPMAP_NEAREST; break;
			case TMIN_NEAREST_PIXEL: minFilter = GL_NEAREST; break;
			case TMIN_NEAREST_PIXEL_NEAREST_MIPMAP: minFilter = GL_NEAREST_MIPMAP_LINEAR; break;
			default: minFilter = GL_NEAREST_MIPMAP_NEAREST;
				ConsoleMessage("unknown minificationFilter of %s",
					tpNode->minificationFilter->strptr);
			}
			else switch (findFieldInTEXTUREMINIFICATIONKEYWORDS(tpNode->minificationFilter->strptr)) {
			case TMIN_AVG_PIXEL:
			case TMIN_AVG_PIXEL_AVG_MIPMAP:
			case TMIN_AVG_PIXEL_NEAREST_MIPMAP:
			case TMIN_DEFAULT:
			case TMIN_FASTEST:
			case TMIN_NICEST:
			case TMIN_NEAREST_PIXEL: minFilter = GL_NEAREST; break;
			case TMIN_NEAREST_PIXEL_NEAREST_MIPMAP: minFilter = GL_LINEAR; break;
			default: minFilter = GL_NEAREST;
				ConsoleMessage("unknown minificationFilter of %s",
					tpNode->minificationFilter->strptr);
			}

			switch (findFieldInTEXTURECOMPRESSIONKEYWORDS(tpNode->textureCompression->strptr)) {
			case TC_DEFAULT: compression = GL_NONE; break;
			case TC_FASTEST: compression = GL_FASTEST; break; /* DEFAULT */
			case TC_HIGH: compression = GL_FASTEST; break;
			case TC_LOW: compression = GL_NONE; break;
			case TC_MEDIUM: compression = GL_NICEST; break;
			case TC_NICEST: compression = GL_NICEST; break;

			default: compression = GL_NEAREST_MIPMAP_NEAREST;
				ConsoleMessage("unknown textureCompression of %s",
					tpNode->textureCompression->strptr);
			}

			BOUNDARY_TO_GL(S);
			BOUNDARY_TO_GL(T);
			BOUNDARY_TO_GL(R);
		}
	}



	if (!haveValidTexturePropertiesNode) {

		/* convert TRUE/FALSE to GL_TRUE/GL_FALSE for wrapS and wrapT */
		Src = Src ? GL_REPEAT : GL_CLAMP_TO_EDGE; //GL_CLAMP;   //du9 changed from CLAMP to CLAMP_TO_EDGE Sept18,2011 to fix panorama seamline visibility
		Trc = Trc ? GL_REPEAT : GL_CLAMP_TO_EDGE; //GL_CLAMP;
		Rrc = Rrc ? GL_REPEAT : GL_CLAMP_TO_EDGE; //GL_CLAMP

		generateMipMaps = GL_TRUE;

		if (me->x > 0 && me->y > 0)
		{
			// experimental code to deal with oblong texture images
			// The 'cause' of blurry oblong textures with GLES2 is the anisotropic
			// scaling of the image needed to square up the image for mipmapping.
			// For example if your texture image is oblong by a factor of 3, then
			// your avatar will need to be 3 times closer to the object to see the
			// same mipmap level as you would with an unscaled image.
			// (OpenGL redbook talks about MipMap levels and rho and lambda.)
			// here we detect if there's big anisotropic scaling/squaring coming up,
			// and if so turn off mipmapping and squaring
			// scene authors need to make their repeating textures (brick, siding,
			// shingles etc) squarish to get mipmapping and avoid moire/scintilation
			float ratio = 1.0f;
			if (me->x < me->y * me->z) ratio = (float)(me->y * me->z) / (float)me->x;
			else ratio = (float)me->x / (float)(me->y * me->z);
			if (ratio > 2.0f) generateMipMaps = GL_FALSE;
		}

		/* choose smaller images to be NEAREST, larger ones to be LINEAR */
		if ((me->x <= 256) || ((me->y * me->z) <= 256)) {
			minFilter = GL_NEAREST_MIPMAP_NEAREST;
			if (!generateMipMaps) minFilter = GL_NEAREST;
			magFilter = GL_NEAREST;
		}
		else {
			minFilter = GL_LINEAR_MIPMAP_NEAREST;
			if (!generateMipMaps) minFilter = GL_LINEAR;
			magFilter = GL_LINEAR;
		}
	}

	//ConsoleMessage ("move_texture_to_opengl cubeFace %x\n",getAppearanceProperties()->cubeFace);

	/* is this a CubeMap? If so, lets try this... */
	//if(0){
	//	printf("before texture to GPU and mipmapping thread=%x time = %lf\n",pthread_self().p,Time1970sec());
	//}


	if (getAppearanceProperties()->cubeFace != 0) {
		//this is a single cubmap face pixeltexture tti (ie from __subTextures in ImageCubemap)
		unsigned char* dest = me->texdata;
		uint32* sp;

		int cx, itype;


		//#if defined (GL_BGRA)
		//iformat = GL_RGBA; format = GL_BGRA;
		//#else
		iformat = GL_RGBA; format = GL_RGBA; itype = GL_UNSIGNED_BYTE;
		if (me->idepthbuffer) {
			iformat = GL_DEPTH_COMPONENT;
			format = GL_DEPTH_COMPONENT;
			itype = GL_FLOAT;
		}
		//#endif


		/* first image in the ComposedCubeMap, do some setups */
		if (getAppearanceProperties()->cubeFace == GL_TEXTURE_CUBE_MAP_POSITIVE_X) {
			FW_GL_TEXPARAMETERI(GL_TEXTURE_CUBE_MAP_EXT, GL_TEXTURE_MIN_FILTER, GL_NEAREST);
			FW_GL_TEXPARAMETERI(GL_TEXTURE_CUBE_MAP_EXT, GL_TEXTURE_MAG_FILTER, GL_NEAREST);
			FW_GL_TEXPARAMETERI(GL_TEXTURE_CUBE_MAP_EXT, GL_TEXTURE_WRAP_S, GL_CLAMP_TO_EDGE);
			FW_GL_TEXPARAMETERI(GL_TEXTURE_CUBE_MAP_EXT, GL_TEXTURE_WRAP_T, GL_CLAMP_TO_EDGE);
			FW_GL_TEXPARAMETERI(GL_TEXTURE_CUBE_MAP_EXT, GL_TEXTURE_WRAP_R, GL_CLAMP_TO_EDGE);
		}

		rx = me->x;
		ry = me->y;
		dest = me->texdata;
		if (1) {
			//flip cubemap textures to be y-down following opengl specs table 3-19
			//'renderman' convention
			//stack method: row chunks at a time
			uint32 tp[512];
			int cy, cyy, icsize;
			sp = (uint32*)me->texdata;
			for (cy = 0; cy < ry / 2; cy++) {
				cyy = ry - cy - 1;
				for (cx = 0; cx < rx; cx += 512) {
					icsize = min(512, rx - cx - 1) * 4;
					memcpy(tp, &sp[cy * rx + cx], icsize);
					memcpy(&sp[cy * rx + cx], &sp[cyy * rx + cx], icsize);
					memcpy(&sp[cyy * rx + cx], tp, icsize);
				}
			}
			//printf("__flipping__\n"); //are we in here on every frame? yes, for generatedcubemaptexture, no for other cubemaps
		}
		generateMipMaps = 0;
		myTexImage2D(generateMipMaps, getAppearanceProperties()->cubeFace, 0, iformat, rx, ry, 0, format, itype, dest);

		/* last thing to do at the end of the setup for the 6th face */
		//if (getAppearanceProperties()->cubeFace == GL_TEXTURE_CUBE_MAP_NEGATIVE_Z) {
		//	glEnable(GL_TEXTURE_CUBE_MAP);
		//	glEnable(GL_TEXTURE_GEN_S);
		//	glEnable(GL_TEXTURE_GEN_T);
		//	glEnable(GL_TEXTURE_GEN_R);
		//}

	}
	else {
		if (me->nodeType == NODE_ImageCubeMapTexture || me->nodeType == NODE_ComposedCubeMapTexture) {
			render_node(me->scenegraphNode);
			return; //June 2022 change, these nodes now load themselves on render_
		}
		if (me->nodeType == NODE_GeneratedCubeMapTexture) {
			me->status = TEX_LOADED;
		}
		//if (me->nodeType == NODE_ImageCubeMapTexture) {
		//	if(me->z == 1){
		//		/* if we have an single 2D image, ImageCubeMap, we have most likely got a png map;
		//		   ________
		//		  |	 T    | - Top
		//		  |L F R B| - Left, Front, Right, Back
		//		  |__D____| - Down(bottom)
		//			let the  render_ImageCubeMapTexture code unpack the maps from this one png */
		//		/* this is ok - what is happening is that we have one image, that needs to be
		//			split up into each face */
		//		/* this should print if we are actually working ok
		//		if (me->status != TEX_LOADED) {
		//			printf ("have ImageCubeMapTexture, but status != TEX_LOADED\n");
		//		}
		//		*/

		//		/* call the routine in Component_CubeMapTexturing.c to split this baby apart */
		//		unpackImageCubeMap(me);
		//		me->status = TEX_LOADED; /* finito */
		//	}else if(me->z == 6){
		//		//likely a .DDS (MS invention) or web3dit (dug9 invention)
		//		//order of images: +x,-x,+y,-y,+z,-z (or R,L,F,B,T,D)
		//		unpackImageCubeMap6(me);
		//		me->status = TEX_LOADED; /* finito */
		//	}
		//	//now the __subTextures individual face textures will show as single faces above
		//} else if(me->nodeType == NODE_GeneratedCubeMapTexture){
		//	//already unpacked into 6 separate PixelTexture tti->texdata during cubemap generation
		//	me->status = TEX_LOADED; /* finito */
		// }
		else {
			int npot;
			/* a pointer to the tex data. We increment the pointer for movie texures */
			mytexdata = me->texdata;
			if (mytexdata == NULL) {
				printf ("mytexdata is null, texture failed, put something here\n");
			}

			glBindTexture (GL_TEXTURE_2D, me->OpenGLTexture);

			/* save this to determine whether we need to do material node
			  within appearance or not */

			/*
			repeatS,repeatT,repeatR
			https://open.gl/textures
			how the texture should be sampled when a coordinate outside the range of 0to 1 is given
			- repeat (*1)
			- mirrored repeat
			- clamp to edge (*2)
			- clamp to border
			subset of opengl supported by web3d:
			http://www.web3d.org/documents/specifications/19775-1/V3.3/Part01/components/texturing.html#Texturecoordinates
			(*1) If repeatS is TRUE (the default), the texture map is repeated outside
				the [0.0, 1.0] texture coordinate range in the S direction so that it fills the shape.
			(*2) If repeatS is FALSE, the texture coordinates are clamped in the S direction to lie
				within the [0.0, 1.0] range.
			*/

			FW_GL_TEXPARAMETERI( GL_TEXTURE_2D, GL_TEXTURE_WRAP_S, Src);
			FW_GL_TEXPARAMETERI( GL_TEXTURE_2D, GL_TEXTURE_WRAP_T, Trc);
#if !defined(GL_ES_VERSION_2_0)
			FW_GL_TEXPARAMETERI( GL_TEXTURE_2D, GL_TEXTURE_WRAP_R, Rrc);
			FW_GL_TEXPARAMETERF(GL_TEXTURE_2D,GL_TEXTURE_PRIORITY, texPri);
			FW_GL_TEXPARAMETERFV(GL_TEXTURE_2D,GL_TEXTURE_BORDER_COLOR,(GLfloat *)&borderColour);
#endif /*  GL_ES_VERSION_2_0 */

#ifdef GL_TEXTURE_MAX_ANISOTROPY_EXT
			FW_GL_TEXPARAMETERF(GL_TEXTURE_2D,GL_TEXTURE_MAX_ANISOTROPY_EXT,anisotropicDegree);
#endif
			iformat = GL_RGBA;
			if (compression != GL_NONE) {
				iformat = GL_COMPRESSED_RGBA;
				//FW_GL_TEXPARAMETERI(GL_TEXTURE_2D, GL_TEXTURE_INTERNAL_FORMAT, GL_COMPRESSED_RGBA);
				//glHint(GL_TEXTURE_COMPRESSION_HINT, compression);
			}
			npot = rdr_caps->av_npot_texture;
			x = me->x;
			y = me->y; // * me->z; //takes care of texture3D using strip image
			z = me->z;

			FW_GL_TEXPARAMETERI( GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, minFilter);
			FW_GL_TEXPARAMETERI( GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, magFilter);
			me->magFilter = magFilter ==  GL_LINEAR ? 1 : 0;  //needed in frag shader for TEX3D simulation of texture3D with texture2D

			/* BGRA is seemingly faster on desktop machines... */
			//#if defined (GL_BGRA)
			//iformat = GL_RGBA; format = GL_BGRA;
			//#else
			format = GL_RGBA;
			//#endif

			/* do the image. */
			if(x && y) {
				unsigned char *dest = mytexdata;

				/* do we have to do power of two textures? */
				if (npot) { //rdr_caps->av_npot_texture) {
					rx = x; ry = y; rz = z;
				} else {
					/* find a power of two that fits */
					static int round_down = 1;
					static int trunc_down = 0;
					rx = 1;
					sx = x;
					while(sx) {sx /= 2; rx *= 2;}
					if(rx/2 == x) {rx /= 2;}
					//if its just a few pixels over a power of 2, round down
					if(round_down)
						if(x < rx && (float)(x - rx/2)/(float)(rx - rx/2) < .25) rx = rx/2;
					//or we might only scale down, never up, if we don't want to malloc again
					if(trunc_down)
						if(x < rx) rx = rx/2;
					ry = 1;
					sy = y;
					while(sy) {sy /= 2; ry *= 2;}
					if(ry/2 == y) {ry /= 2;}
					if(round_down)
						if(y < ry && (float)(y - ry/2)/(float)(ry - ry/2) < .25) ry = ry/2;
					if(trunc_down)
						if(y < ry) ry = ry/2;

					rz = 1;
					sz = z;
					while(sz) {sz /= 2; rz *= 2;}
					if(rz/2 == z) {rz /= 2;}
					if(round_down)
						if(z < rz && (float)(z - rz/2)/(float)(rz - rz/2) < .25) rz = rz/2;
					if(trunc_down)
						if(z < rz) rz = rz/2;
				}

				if (gglobal()->internalc.global_print_opengl_errors) {
					DEBUG_MSG("initial texture scale to %d %d\n",rx,ry);
				}

				//ConsoleMessage ("loadTextureNode, runtime texture size %d",gglobal()->display.rdr_caps.runtime_max_texture_size);
				if(z > 1){
					//its a texture3D / volume image
					int emulating3D_TILED;
					emulating3D_TILED = TRUE;
					generateMipMaps = FALSE;
					if(emulating3D_TILED){
						//tiled uses more of the max_texture_size x max_texture_size
						//	texture3D emulator via TILED texture2D
						//  reason for emulating: 2016 GLES2 via ANGLEPROJECT(gles emulator over DirectX on windows)
						//     doesn't have Texture3D or Texture3DOES or Texture3DEXT.
						//  reason for TILES: an oblong Y-STRIP approach exceded max texture size in Y (but had lots left in X)
						//     desktop computer max_size (of 2D image in one dimension) 16384
						//     android phone max_size 4096
						//     and so would be resampled (blurry) in y and good in x
						//     using tiles means room for more full z slices ie 256x256x256 == 4096x4096 == 16M,
						//			512x512x512 == 134M == 16384x16384/2, and therefore less blurry images
						//  tiles start in upper left with z=0, increase in y,
						//  then when hit ny tiles in a y strip, move right one tile, and restart at top
						//  uniform tex3dTiles[3] = {nx,ny,z}
						//  example ny = 4, nx = 3, z = 11
						//  1  5  9
						//  2  6  10
						//  3  7  11
						//  4  8
						//
						//
						int rc,sc,c,cube_root, max_size;
						unsigned char *texdataTiles = NULL;
						uint32 *p2, *p1;
						int nx, ny, ix, iy, nxx, nyy;
						int iz,j,k;

						max_size = rdr_caps->runtime_max_texture_size;
						//if(32bit) I find process doesn't have enough RAM left for opengl to malloc 512x512x512x4byte.
						//could try single channel, single byte textures, but for now we'll keep it under 17M pixels
						if(x * y * z > 256 * 256 * 256)
							max_size = min(max_size,4096);
						//max_size = 2048; //can re-set here for experiments
						cube_root = (int)pow( max_size * max_size  + 3, 1.0/3.0);
						c = cube_root;
						//need lower-power-of-two so we squeeze into space available, and leave a little
						rc = 1;
						sc = c;
						while(sc) {sc /= 2; rc *= 2;}
						if(rc > c) {rc /= 2;}
						//ConsoleMessage("pow2 cube root %d\n",rc);
						cube_root = rc;
						if(rx != x || ry != y || rz != z || rx > cube_root || ry > cube_root || rz > cube_root) {
							/* do we have texture limits???
							dug9: windows intel i5: desktop opengl and uwp/angleproject 16384
							16384 x 16394 = 268M. cube-root 268M = 645.xx lets round down to pow2: 512
							android LG nexus 4096
							4096 x 4096 = 16.7M; cube-root 16.7M = 256.
							*/
							if (rx > cube_root) rx = cube_root;
							if (ry > cube_root) ry = cube_root;
							if (rz > cube_root) rz = cube_root;
						}

						if (gglobal()->internalc.global_print_opengl_errors) {
							DEBUG_MSG("texture size after maxTextureSize taken into account: %d %d, from %d %d\n",rx,ry,x,y);
						}
						//ConsoleMessage("texture size after maxTextureSize taken into account: %d %d %d, from %d %d %d\n",rx,ry,rz,x,y,z);

						//rescale sub-images if/as needed
						dest = mytexdata;
						//if(rx != x || ry != y || rz != z){
						//if(rx > x || ry > y || rz > z){
						if(x > rx || y > ry || z > rz){
							int mx,my,mz;
							mx = min(x,rx);
							my = min(y,ry);
							mz = min(z,rz);
							dest = MALLOC(unsigned char *, 4 * rx * ry * rz);
							myScaleImage3D(x,y,z,mx,my,mz,mytexdata,dest);
							x = mx;
							y = my;
							z = mz;
							FREE_IF_NZ(me->texdata);
						}

						//COMPUTE GRADIENT - we'll do unconditionally if channels == 1 for 3D image
						//and hope that the one info channel is alpha because we overwrite rgb
						if(me->channels == 1){
							//alpha only scalar image, RGB are free to hold gradient
							compute_3D_alpha_gradient_store_rgb((char *)dest,x,y,z);
						}

						ny = (int) sqrt(z+1);
						nx = z / ny;
						nx = z - nx*ny > 0 ? nx+1 : nx;

						me->tiles[0] = nx; //let the shader tiled emulator for texture3D know via uniform about the tile layout
						me->tiles[1] = ny;
						me->tiles[2] = z;
						//ConsoleMessage("Tiles ny %d nx %d zplanes %d\n",nx,ny,z);
						nxx = nx*rx;
						nyy = ny*ry;

						//place in tile formation - a series of fullish y strips
						texdataTiles =  MALLOC(unsigned char *,nxx * nyy * 4);
						p2 = (uint32 *)texdataTiles;
						p1 = (uint32 *)dest;
						for(iz=0;iz<z;iz++){
							iy = iz % ny;
							ix = iz / ny;
							for(j=0;j<y;j++){
								for(k=0;k<x;k++){
									int ifrom, ito;
									uint32 pixel;
									ifrom = (iz*y + j)*x + k;
									ito = (iy*y + j)*nxx + (ix*x) + k;
									pixel = p1[ifrom];
									p2[ito] = pixel;
								}
							}
						}
						if(0){
							//write out tiled image for inspection
							textureTableIndexStruct_s tti2, *tti3;
							tti3 = &tti2;
							tti3->x = nxx;
							tti3->y = nyy;
							tti3->z = 1;
							tti3->channels = 4;
							tti3->texdata = texdataTiles;
							saveImage_web3dit(tti3, "test_tiled_texture.web3dit");
						}

						myTexImage2D(generateMipMaps, GL_TEXTURE_2D, 0, iformat,  nxx, nyy, 0, format, GL_UNSIGNED_BYTE, texdataTiles);
						ConsoleMessage("final texture2D size %d %d\n",nxx,nyy);
						FREE_IF_NZ(texdataTiles);
						if(dest != me->texdata) FREE_IF_NZ(dest);
					}else{
						//use Texture3D which android and winRT/uwp don't have
					}
				}else{
					//ordinary 2D image textures
					if(rx != x || ry != y || rx > rdr_caps->runtime_max_texture_size || ry > rdr_caps->runtime_max_texture_size) {
						/* do we have texture limits???
						dug9: windows intel i5: desktop opengl and uwp/angleproject 16384
						16384 x 16394 = 268M. cube-root 268M = 645.xx lets round down to pow2: 512
						android LG nexus 4096
						4096 x 4096 = 16.7M; cube-root 16.7M = 256.
						*/
						rx = min(rx,rdr_caps->runtime_max_texture_size);
						ry = min(ry,rdr_caps->runtime_max_texture_size);
					}

					if (gglobal()->internalc.global_print_opengl_errors) {
						DEBUG_MSG("texture size after maxTextureSize taken into account: %d %d, from %d %d\n",rx,ry,x,y);
					}
					//ConsoleMessage("texture size after maxTextureSize taken into account: %d %d, from %d %d\n",rx,ry,x,y);

					/* it is a power of 2, lets make sure it is square */
					/* ES 2.0 needs this for cross-platform; do not need to do this for desktops, but
					   lets just keep things consistent
					   But if not mipmapping, then (experience with win32 GLES2 emulator and QNX device)
					   then it's not necessary to square the image, although current code will get here with
					   generateMipMap always true.
					   */
					if (rx != ry) {
						if(generateMipMaps){
							if (rx>ry)ry=rx;
							else rx=ry;
						}
					}
					/* if scaling is ok... */
					if ((x==rx) && (y==ry)) {
						dest = mytexdata;
					} else {

						/* try this texture on for size, keep scaling down until we can do it */
						/* all textures are 4 bytes/pixel */
						size_t total_size = (size_t)4L * rx * ry;
						dest = MALLOC(unsigned char *, total_size); //4 * rx * ry);

						myScaleImage(x,y,rx,ry,mytexdata,dest);
					}
					if(rx > 8192 || ry > 8192) ConsoleMessage("texture size rx %d ry %d\n",rx,ry);
					myTexImage2D(generateMipMaps, GL_TEXTURE_2D, 0, iformat,  rx, ry, 0, format, GL_UNSIGNED_BYTE, dest);
				}
				if(mytexdata != dest) {
					FREE_IF_NZ(dest);
				}
			}

			/* we can get rid of the original texture data here */
			FREE_IF_NZ (me->texdata);
		}
	}


	/* ensure this data is written to the driver for the rendering context */
	FW_GL_FLUSH();

	/* and, now, the Texture is loaded */
	me->status = TEX_LOADED;

	#ifdef PATH_PLANNER
	pthread_mutex_unlock( &gl_mutex1 );
	#endif //PATH_PLANNER

}


/**********************************************************************************
 bind the image,

	itype 	tells us whether it is a PixelTexture, ImageTexture or MovieTexture.

	parenturl  is a pointer to the url of the parent (for relative loads) OR
		a pointer to the image data (PixelTextures only)

	url	the list of urls from the VRML file, or NULL (for PixelTextures)

	texture_num	the OpenGL texture identifier

	repeatS, repeatT VRML fields

	param - vrml fields, but translated into GL_TEXTURE_ENV_MODE, GL_MODULATE, etc.
************************************************************************************/

void new_bind_image(struct X3D_Node *node, struct multiTexParams *param) {
	int thisTexture;
	int thisTextureType;
	struct X3D_ImageTexture *it;
	struct X3D_PixelTexture *pt;
	struct X3D_BufferTexture* bt;
	struct X3D_MovieTexture *mt;
	struct X3D_ImageCubeMapTexture *ict;
	struct X3D_GeneratedCubeMapTexture *gct;
	struct X3D_GeneratedTexture* gt;

	textureTableIndexStruct_s *myTableIndex;
	//float dcol[] = {0.8f, 0.8f, 0.8f, 1.0f};
	ppTextures p;
	struct Multi_String *mfurl = NULL;
	ttglobal tg = gglobal();
	p = (ppTextures)tg->Textures.prv;
	//#define DEBUG_TEX ConsoleMessage

//	GET_THIS_TEXTURE;
//#define GET_THIS_TEXTURE
	thisTextureType = node->_nodeType;
	if (thisTextureType==NODE_ImageTexture){
		it = (struct X3D_ImageTexture*) node;
		mfurl = &it->url;
		thisTexture = it->__textureTableIndex;
	} else if (thisTextureType==NODE_PixelTexture){
		pt = (struct X3D_PixelTexture*) node;
		thisTexture = pt->__textureTableIndex;
	} else if (thisTextureType == NODE_BufferTexture) {
		bt = (struct X3D_BufferTexture*)node;
		thisTexture = bt->__textureTableIndex;
	} else if (thisTextureType == NODE_GeneratedTexture) {
		gt = (struct X3D_GeneratedTexture*)node;
		thisTexture = gt->__textureTableIndex;
	} else if (thisTextureType==NODE_MovieTexture){
		mt = (struct X3D_MovieTexture*) node;
		thisTexture = mt->__textureTableIndex;
		mfurl = &mt->url;
	} else if (thisTextureType == NODE_ComposedCubeMapTexture) {
		ict = (struct X3D_ImageCubeMapTexture*)node;
		thisTexture = ict->__textureTableIndex;
		mfurl = &ict->url;
	} else if (thisTextureType==NODE_ImageCubeMapTexture){
		ict = (struct X3D_ImageCubeMapTexture*) node;
		thisTexture = ict->__textureTableIndex;
		mfurl = &ict->url;
	} else if (thisTextureType==NODE_GeneratedCubeMapTexture){
		gct = (struct X3D_GeneratedCubeMapTexture*) node;
		thisTexture = gct->__textureTableIndex;
	} else if (thisTextureType==NODE_PixelTexture3D){
		struct X3D_PixelTexture3D *pt3d;
		pt3d = (struct X3D_PixelTexture3D*) node;
		thisTexture = pt3d->__textureTableIndex;
	} else if (thisTextureType==NODE_ImageTexture3D){
		struct X3D_ImageTexture3D *pt3d;
		pt3d = (struct X3D_ImageTexture3D*) node;
		thisTexture = pt3d->__textureTableIndex;
		mfurl = &pt3d->url;
	} else if (thisTextureType==NODE_ComposedTexture3D){
		struct X3D_ComposedTexture3D *pt3d;
		pt3d = (struct X3D_ComposedTexture3D*) node;
		thisTexture = pt3d->__textureTableIndex;
	} else {
		ConsoleMessage ("Invalid type for texture, %s\n",stringNodeType(thisTextureType));
		return;
	}

	myTableIndex = getTableIndex(thisTexture);
	if (myTableIndex->status != TEX_LOADED) {
		DEBUG_TEX("new_bind_image, I am %p, textureStackTop %d, thisTexture is %d myTableIndex %p status %s\n",
		node,tg->RenderFuncs.textureStackTop,thisTexture,myTableIndex, texst(myTableIndex->status));
		//printf ("new_bind_image (%d), I am %p, textureStackTop %d, thisTexture is %d myTableIndex %p status %s\n",
		//__LINE__,node,tg->RenderFuncs.textureStackTop,thisTexture,myTableIndex, texst(myTableIndex->status));
	}

	/* default here; this is just a blank texture */
	tg->RenderFuncs.boundTextureStack[tg->RenderFuncs.textureStackTop] = tg->Textures.defaultBlankTexture;
	switch (myTableIndex->status) {
		case TEX_NOTLOADED:
			DEBUG_TEX("feeding texture %p to texture thread...\n", myTableIndex);
			if(mfurl && mfurl->n == 0) {
				//for <ImageTexture /> with url not declared, we should get the default blank image
				myTableIndex->status = TEX_NEEDSBINDING;
			} else {
				myTableIndex->status = TEX_LOADING;
				send_texture_to_loader(myTableIndex);
			}
			break;

		case TEX_LOADING:
		case TEX_READ:
			DEBUG_TEX("I've to wait for %p...\n", myTableIndex);
			break;

		case TEX_NEEDSBINDING:
			DEBUG_TEX("texture loaded into memory... now lets load it into OpenGL...\n");
			if (myTableIndex->no_gl == 0) {
				move_texture_to_opengl(myTableIndex);

				// do always #ifdef PATH_PLANNER
				// JAS - skipping a rendering loop
				// if we are ok, go direct to rendering this texture.
				if (myTableIndex->status != TEX_LOADED) {
					printf("issue going from TEX_NEEDSBINDING to TEX_LOADED, is %s\n",
						texst(myTableIndex->status));
					break;
				}
				// do always #else
				// do always break;
				// do always #endif  //PATH_PLANNER
			}

		case TEX_LOADED:
			//DEBUG_TEX("now binding to pre-bound tex %u\n", myTableIndex->OpenGLTexture);

			/* set the texture depth - required for Material diffuseColor selection */
			if (myTableIndex->hasAlpha) tg->RenderFuncs.last_texture_type =  TEXTURE_ALPHA;
			else tg->RenderFuncs.last_texture_type = TEXTURE_NO_ALPHA;

//printf ("last_texture_type = TEXTURE_NO_ALPHA now\n"); last_texture_type=TEXTURE_NO_ALPHA;

			if (myTableIndex->OpenGLTexture == TEXTURE_INVALID) {

				DEBUG_TEX("no openGLtexture here status %s\n", texst(myTableIndex->status));
				return;
			}

			tg->RenderFuncs.boundTextureStack[tg->RenderFuncs.textureStackTop] = myTableIndex->OpenGLTexture;
            //printf ("new_bind, boundTextureStack[%d] set to %d\n",tg->RenderFuncs.textureStackTop,myTableIndex->OpenGLTexture);


			/* save the texture params for when we go through the MultiTexture stack. Non
			   MultiTextures should have this textureStackTop as 0 */

			if (param != NULL) {
				struct multiTexParams *textureParameterStack = (struct multiTexParams *) tg->RenderTextures.textureParameterStack;
				memcpy(&(textureParameterStack[tg->RenderFuncs.textureStackTop]), param,sizeof (struct multiTexParams));
				//memcpy(&(tg->RenderTextures.textureParameterStack[tg->RenderFuncs.textureStackTop]), param,sizeof (struct multiTexParams));
			}
			p->textureInProcess = -1; /* we have finished the whole process */
			break;

		case TEX_UNSQUASHED:
		default: {
			printf ("unknown texture status %d\n",myTableIndex->status);
		}
	}
	//#define DEBUG_TEX
}

int get_bound_image(struct X3D_Node *node) {
	int thisTexture;
	int thisTextureType;
	struct X3D_ImageTexture *it;
	struct X3D_PixelTexture *pt;
	struct X3D_BufferTexture* bt;
	struct X3D_MovieTexture *mt;
	struct X3D_ImageCubeMapTexture *ict;
	struct X3D_GeneratedCubeMapTexture *gct;
	struct X3D_GeneratedTexture* gt;

	textureTableIndexStruct_s *myTableIndex;
	//float dcol[] = {0.8f, 0.8f, 0.8f, 1.0f};
	ppTextures p;
	struct Multi_String *mfurl = NULL;
	ttglobal tg = gglobal();
	p = (ppTextures)tg->Textures.prv;
	//#define DEBUG_TEX ConsoleMessage

//	GET_THIS_TEXTURE;
//#define GET_THIS_TEXTURE
	thisTextureType = node->_nodeType;
	if (thisTextureType==NODE_ImageTexture){
		it = (struct X3D_ImageTexture*) node;
		mfurl = &it->url;
		thisTexture = it->__textureTableIndex;
	} else if (thisTextureType==NODE_PixelTexture){
		pt = (struct X3D_PixelTexture*) node;
		thisTexture = pt->__textureTableIndex;
	} else if (thisTextureType == NODE_BufferTexture) {
		bt = (struct X3D_BufferTexture*)node;
		thisTexture = bt->__textureTableIndex;
	} else if (thisTextureType == NODE_GeneratedTexture) {
		gt = (struct X3D_GeneratedTexture*)node;
		thisTexture = gt->__textureTableIndex;
	} else if (thisTextureType==NODE_MovieTexture){
		mt = (struct X3D_MovieTexture*) node;
		thisTexture = mt->__textureTableIndex;
		mfurl = &mt->url;
	} else if (thisTextureType == NODE_ComposedCubeMapTexture) {
		ict = (struct X3D_ImageCubeMapTexture*)node;
		thisTexture = ict->__textureTableIndex;
		mfurl = &ict->url;
	} else if (thisTextureType==NODE_ImageCubeMapTexture){
		ict = (struct X3D_ImageCubeMapTexture*) node;
		thisTexture = ict->__textureTableIndex;
		mfurl = &ict->url;
	} else if (thisTextureType==NODE_GeneratedCubeMapTexture){
		gct = (struct X3D_GeneratedCubeMapTexture*) node;
		thisTexture = gct->__textureTableIndex;
	} else if (thisTextureType==NODE_PixelTexture3D){
		struct X3D_PixelTexture3D *pt3d;
		pt3d = (struct X3D_PixelTexture3D*) node;
		thisTexture = pt3d->__textureTableIndex;
	} else if (thisTextureType==NODE_ImageTexture3D){
		struct X3D_ImageTexture3D *pt3d;
		pt3d = (struct X3D_ImageTexture3D*) node;
		thisTexture = pt3d->__textureTableIndex;
		mfurl = &pt3d->url;
	} else if (thisTextureType==NODE_ComposedTexture3D){
		struct X3D_ComposedTexture3D *pt3d;
		pt3d = (struct X3D_ComposedTexture3D*) node;
		thisTexture = pt3d->__textureTableIndex;
	} else {
		ConsoleMessage ("Invalid type for texture, %s\n",stringNodeType(thisTextureType));
		return -1;
	}

	myTableIndex = getTableIndex(thisTexture);
	return myTableIndex->OpenGLTexture;
}
