/*

  FreeWRL support library.
  Scenegraph rendering.

*/

/****************************************************************************
    This file is part of the FreeWRL/FreeX3D Distribution.

    Copyright 2009 CRC Canada. (http://www.crc.gc.ca)

    FreeWRL/FreeX3D is free software: you can redistribute it and/or modify
    it under the terms of the GNU Lesser Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    FreeWRL/FreeX3D is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with FreeWRL/FreeX3D.  If not, see <http://www.gnu.org/licenses/>.
****************************************************************************/



#include <config.h>
#include <system.h>
#include <display.h>
#include <internal.h>

#include <libFreeWRL.h>

#include "../vrml_parser/Structs.h"
#include "../main/headers.h"
#include "../scenegraph/Component_ProgrammableShaders.h"

#include "Polyrep.h"
#include "Collision.h"
#include "../scenegraph/quaternion.h"
#include "Viewer.h"
#include "LinearAlgebra.h"
#include "../input/SensInterps.h"
#include "system_threads.h"
#include "threads.h"

#include "../opengl/OpenGL_Utils.h"
#include "../opengl/Textures.h"
#include "../scenegraph/Component_Shape.h"
#include "RenderFuncs.h"
#include "../ui/common.h"

typedef float shaderVec4[4];


struct profile_entry {
	char *name;
	double start;
	double accum;
	int hits;
};

struct point_XYZ3 {
	struct point_XYZ p1;
	struct point_XYZ p2;
	struct point_XYZ p3;
};

typedef struct {
int calltype;
union {
	struct arrays {
		int arrays_mode;
		int arrays_count;
		int arrays_first;
	} arrays;
	struct elements {
		int elements_mode;
		int elements_count;
		GLenum elements_type;
		void *elements_indices;
	} elements;
};
} draw_call_params;

typedef struct pRenderFuncs{
	int profile_entry_count;
	struct profile_entry profile_entries[100];
	int profiling_on;

	//int cur_hits;//=0;
	void *empty_group;//=0;
	//struct point_XYZ ht1, ht2; not used
	struct point_XYZ hyper_r1,hyper_r2; /* Transformed ray for the hypersensitive node */
	struct currayhit rayph;
	struct X3D_Node *rootNode;//=NULL;	/* scene graph root node */
	struct Vector *libraries; //vector of extern proto library scenes in X3D_Proto format that are parsed shallow (not instanced scenes) - the library protos will be in X3D_Proto->protoDeclares vector
	struct X3D_Anchor *AnchorsAnchor;// = NULL;
	struct currayhit rayHit; //,rayHitHyper;
	//struct trenderstate renderstate;
	Stack* renderstate;
	int renderLevel;

	// which Shader is currently in use?
	GLint currentShader;
	Stack *render_geom_stack;
	Stack *sensor_stack;
	Stack *ray_stack;
	Stack *shaderflags_stack;
	Stack *fog_stack;
	Stack* ectx_stack; //executionContext

	//struct point_XYZ t_r1,t_r2,t_r3; /* transformed ray */
	struct point_XYZ3 t_r123;
	struct point_XYZ hp;
	Stack *usehits_stack;
	Stack *usehitsB_stack;
	Stack *pickablegroupdata_stack;
	Stack *draw_call_params_stack;
}* ppRenderFuncs;
void *RenderFuncs_constructor(){
	void *v = MALLOCV(sizeof(struct pRenderFuncs));
	memset(v,0,sizeof(struct pRenderFuncs));
	return v;
}
void RenderFuncs_init(struct tRenderFuncs *t){
	//public

	t->BrowserAction = FALSE;
	//	t->hitPointDist; /* distance in ray: 0 = r1, 1 = r2, 2 = 2*r2-r1... */
	///* used to save rayhit and hyperhit for later use by C functions */
	//t->hyp_save_posn;
	//t->hyp_save_norm;t->ray_save_posn;
	t->hypersensitive = 0;
	t->hyperhit = 0;
	t->have_transparency=FALSE;/* did any Shape have transparent material? */
	/* material node usage depends on texture depth; if rgb (depth1) we blend color field
	   and diffusecolor with texture, else, we dont bother with material colors */
	t->last_texture_type = NOTEXTURE;

	//private
	t->prv = RenderFuncs_constructor();
	{
		ppRenderFuncs p = (ppRenderFuncs)t->prv;
		p->profile_entry_count = 0;
		p->profiling_on = 0; //toggle on with '.' on keyboard

		p->empty_group=0;
		p->rootNode=NULL;	/* scene graph root node */
		p->libraries=newVector(void3 *,1);
		p->AnchorsAnchor = NULL;
		t->rayHit = (void *)&p->rayHit;
		//t->rayHitHyper = (void *)&p->rayHitHyper;
		p->renderLevel = 0;
		p->renderstate = newStack(struct trenderstate);
		struct trenderstate ttr;
		memset(&ttr, 0, sizeof(struct trenderstate));
		stack_push(struct trenderstate, p->renderstate, ttr);
		p->render_geom_stack = newStack(int);
		p->sensor_stack = newStack(struct currayhit);
		p->ray_stack = newStack(struct point_XYZ3);
		p->usehits_stack = newStack(usehit);
		p->usehitsB_stack = newStack(usehit);
		p->pickablegroupdata_stack = newStack(void*);
		p->shaderflags_stack = newStack(shaderflagsstruct); //newStack(unsigned int);
		p->fog_stack = newStack(struct X3D_Node*);
		p->ectx_stack = newStack(struct X3D_Node*);
		p->draw_call_params_stack = newStack(draw_call_params);
		//t->t_r123 = (void *)&p->t_r123;
		t->hp = (void *)&p->hp;
	}

	//setLightType(HEADLIGHT_LIGHT,2); // ensure that this is a DirectionalLight.
}
//the following usehit functions are for node-node scenarios such as picksensor and transformsensor
//more precisely for node_USE-node_USE aka USE_USE scenarios 
//when a node is 'rendered' if its VF_USE flag is set, then we call usehit_add(self,modelviewmatrix)
//then in do_first() > do_activity() the use-use combinations are each applied.
//at least one in the use-use pair needs to hold a pointer to the other node to use as a lookup 
void usehit_add(struct X3D_Node * node, double *modelviewmatrix){
	//called from render_hier when/each-use-time a VF_USE node is hit
	usehit uhit;
	ppRenderFuncs p = (ppRenderFuncs)gglobal()->RenderFuncs.prv;
	uhit.node = node;
	memcpy(uhit.mvm,modelviewmatrix,16*sizeof(double)); //deep copy
	uhit.userdata = NULL;
	vector_pushBack(usehit,p->usehits_stack,uhit);  //fat elements do another deep copy
}
void usehit_add2(struct X3D_Node * node, double *modelviewmatrix, void *userdata){
	//called from render_hier when/each-use-time a VF_USE node is hit
	usehit uhit;
	ppRenderFuncs p = (ppRenderFuncs)gglobal()->RenderFuncs.prv;
	uhit.node = node;
	memcpy(uhit.mvm,modelviewmatrix,16*sizeof(double)); //deep copy
	uhit.userdata = userdata;
	vector_pushBack(usehit,p->usehits_stack,uhit);  //fat elements do another deep copy
}
usehit * usehit_next(struct X3D_Node *node, usehit *lasthit){
	//called from do_first() > do_activity() when one of the use-use pair is searching for another of its mates
	//call with lasthit = NULL the first time, and otherwise the previous hit to continue searching
	int i, istart;
	usehit *ret, *item;
	ppRenderFuncs p = (ppRenderFuncs)gglobal()->RenderFuncs.prv;
	ret = NULL;
	if(vectorSize(p->usehits_stack)>0){
		//find lasthit
		istart = 0;
		if(lasthit) {
			//size_t size;
			//ptrdiff_t delta;
			//void *start;
			//size = sizeof(usehit);
			//start = (char*)vector_get_ptr(usehit,p->usehits_stack,0);
			//delta = (char*)lasthit - (char*)start;
			//istart = delta/size + 1;
			istart = ((char*)lasthit - (char*)vector_get_ptr(usehit,p->usehits_stack,0))/sizeof(usehit) + 1;
		}
		//search starting at lasthit+1
		for(i=istart;i<p->usehits_stack->n;i++){
			item = vector_get_ptr(usehit,p->usehits_stack,i);
			if(item->node == node){
				ret = item;
				break;
			}
		}
	}
	return ret; //returing pointer to p->usehits fat element
}
void usehit_clear(){
	//called at the end of do_first (once per frame, after USE_USE pairing and action, and before rendering) 
	//to clear all the USE hits from last frame
	ppRenderFuncs p = (ppRenderFuncs)gglobal()->RenderFuncs.prv;
	p->usehits_stack->n = 0;
}

//USEHITB - for Component_Picking.c when traversing the sub-scenegraph
// to get geometry nodes, and their transform, and any more PickingGroup userdata
// (sorry - I just copied the useHit functions above, normally I refactor but short on time -dug9 dec31,2016)
void usehitB_add(struct X3D_Node * node, double *modelviewmatrix){
	//called from render_hier when/each-use-time a VF_USE node is hit
	usehit uhit;
	ppRenderFuncs p = (ppRenderFuncs)gglobal()->RenderFuncs.prv;
	uhit.node = node;
	memcpy(uhit.mvm,modelviewmatrix,16*sizeof(double)); //deep copy
	uhit.userdata = NULL;
	vector_pushBack(usehit,p->usehitsB_stack,uhit);  //fat elements do another deep copy
}
void usehitB_add2(struct X3D_Node * node, double *modelviewmatrix, void *userdata){
	//called from render_hier when/each-use-time a VF_USE node is hit
	usehit uhit;
	ppRenderFuncs p = (ppRenderFuncs)gglobal()->RenderFuncs.prv;
	uhit.node = node;
	memcpy(uhit.mvm,modelviewmatrix,16*sizeof(double)); //deep copy
	uhit.userdata = userdata;
	vector_pushBack(usehit,p->usehitsB_stack,uhit);  //fat elements do another deep copy
}
usehit * usehitB_next(struct X3D_Node *node, usehit *lasthit){
	//called from do_first() > do_activity() when one of the use-use pair is searching for another of its mates
	//call with lasthit = NULL the first time, and otherwise the previous hit to continue searching
	int i, istart;
	usehit *ret, *item;
	ppRenderFuncs p = (ppRenderFuncs)gglobal()->RenderFuncs.prv;
	ret = NULL;
	if(vectorSize(p->usehitsB_stack)>0){
		//find lasthit
		istart = 0;
		if(lasthit) {
			//size_t size;
			//ptrdiff_t delta;
			//void *start;
			//size = sizeof(usehit);
			//start = (char*)vector_get_ptr(usehit,p->usehits_stack,0);
			//delta = (char*)lasthit - (char*)start;
			//istart = delta/size + 1;
			istart = ((char*)lasthit - (char*)vector_get_ptr(usehit,p->usehitsB_stack,0))/sizeof(usehit) + 1;
		}
		//search starting at lasthit+1
		for(i=istart;i<p->usehitsB_stack->n;i++){
			item = vector_get_ptr(usehit,p->usehitsB_stack,i);
			if(item->node == node){
				ret = item;
				break;
			}
		}
	}
	return ret; //returing pointer to p->usehits fat element
}
void usehitB_clear(){
	//called at the end of do_first (once per frame, after USE_USE pairing and action, and before rendering) 
	//to clear all the USE hits from last frame
	ppRenderFuncs p = (ppRenderFuncs)gglobal()->RenderFuncs.prv;
	p->usehitsB_stack->n = 0;
}
Stack *getUseHitBStack(){
	ppRenderFuncs p = (ppRenderFuncs)gglobal()->RenderFuncs.prv;
	return p->usehitsB_stack;
}

//PickableGroup can be several parents above a usehit picktarget node
//see prep_PickableGroup, fin_PickableGroup for push and pop, 
//see below for call to getpickablegroupdata() in render
void push_pickablegroupdata(void *userdata){
	ppRenderFuncs p = (ppRenderFuncs)gglobal()->RenderFuncs.prv;
	stack_push(void*,p->pickablegroupdata_stack,userdata);
}
void pop_pickablegroupdata(){
	ppRenderFuncs p = (ppRenderFuncs)gglobal()->RenderFuncs.prv;
	stack_pop(void*,p->pickablegroupdata_stack);
}
void *getpickablegroupdata(){
	void *ret;
	ppRenderFuncs p = (ppRenderFuncs)gglobal()->RenderFuncs.prv;
	ret = NULL;
	if(vectorSize(p->pickablegroupdata_stack)>0)
		ret = stack_top(void*,p->pickablegroupdata_stack);
	return ret;
}


void unload_libraryscenes();
int gc_broto_instance(struct X3D_Proto* node);
void RenderFuncs_clear(struct tRenderFuncs *t){
	ppRenderFuncs p = (ppRenderFuncs)t->prv;
	unload_libraryscenes();
	deleteVector(void3 *,p->libraries);
	deleteVector(int,p->render_geom_stack);
	deleteVector(struct currayhit,p->sensor_stack);
	deleteVector(struct point_XYZ3,p->ray_stack);
	deleteVector(usehit,p->usehits_stack);
	//deleteVector(unsigned int,p->shaderflags_stack);
	deleteVector(shaderflagsstruct,p->shaderflags_stack);
	deleteVector(struct X3D_Node*,p->fog_stack);
	deleteVector(draw_call_params,p->draw_call_params_stack);
}
void unload_libraryscenes(){
	ppRenderFuncs p = (ppRenderFuncs)gglobal()->RenderFuncs.prv;
	//freeing these library scenes should be done during exit procedures before gglobal gc, perhaps in 
	// finalizeRenderSceneUpdateScene
	// or perhaps when changing scenes. Perhaps libraries should be in a Scene context. 
	// One old idea not implemented: all scenes should first be parsed to libraryScene (nothing registered, empty protoInstance bodies)
	// then scene instanced like a proto. That would speed up Anchoring between scene files ie between rooms. 
	// (Avatar state would be carried between scenes in browser key,value attributes like metadata
	if(p->libraries){
		int i;
		for(i=0;i<vectorSize(p->libraries);i++){
			struct X3D_Proto *libscn;
			char *url;
			void3 *ul;
			ul = vector_get(struct void3*,p->libraries,i);
			if(ul){
				url = (char *)ul->one;
				libscn = (struct X3D_Proto*) ul->two;
				//unload_broto(libscn); //nothing to un-register - library scenes aren't registered
				gc_broto_instance(libscn);
				deleteVector(struct X3D_Node*,libscn->_parentVector);
				freeMallocedNodeFields((struct X3D_Node*)libscn);
				FREE_IF_NZ(libscn);
				FREE_IF_NZ(url);
				FREE_IF_NZ(ul);
				//FREE_IF_NZ(res);
				vector_set(struct void3*,p->libraries,i,NULL);
			}
		}
		p->libraries->n = 0;
	}
}

void transformPositionToEye(float *pos)
{
	int i;
    GLDOUBLE modelMatrix[16], *b;
	float *a;
	float aux[4];
    FW_GL_GETDOUBLEV(GL_MODELVIEW_MATRIX, modelMatrix);

    /* pre-multiply the light position, as per the orange book, page 216,
     "OpenGL specifies that light positions are transformed by the modelview
     matrix when they are provided to OpenGL..." */
    /* DirectionalLight?  PointLight, SpotLight? */

	// assumes pos[3] = 0.0; only use first 3 of these numbers
	transformf(aux,pos,modelMatrix);

	for(i=0;i<3;i++){
		pos[i] = aux[i];
	}
}

void transformDirectionToEye(float *dir)
{
	int i;
    GLDOUBLE modelMatrix[16], *b;
	float *a;
	float aux[4];
    FW_GL_GETDOUBLEV(GL_MODELVIEW_MATRIX, modelMatrix);

	b = modelMatrix;
	a = dir;
	aux[0] = (float) (b[0]*a[0] +b[4]*a[1] +b[8]*a[2] );
	aux[1] = (float) (b[1]*a[0] +b[5]*a[1] +b[9]*a[2] );
	aux[2] = (float) (b[2]*a[0] +b[6]*a[1] +b[10]*a[2]);
	for(i=0;i<3;i++)
		dir[i] = aux[i];
}

void transformLightToEye(float *pos, float* dir)
{
	int i;
    GLDOUBLE modelMatrix[16], *b;
	float *a;
	shaderVec4 aux, auxt;
    FW_GL_GETDOUBLEV(GL_MODELVIEW_MATRIX, modelMatrix);

/*
ConsoleMessage ("nvm %4.2f %4.2f %4.2f %4.2f",modelMatrix[0],modelMatrix[1],modelMatrix[2],modelMatrix[3]);
ConsoleMessage ("nvm %4.2f %4.2f %4.2f %4.2f",modelMatrix[4],modelMatrix[5],modelMatrix[6],modelMatrix[7]);
ConsoleMessage ("nvm %4.2f %4.2f %4.2f %4.2f",modelMatrix[8],modelMatrix[9],modelMatrix[10],modelMatrix[11]);
ConsoleMessage ("nvm %4.2f %4.2f %4.2f %4.2f",modelMatrix[12],modelMatrix[13],modelMatrix[14],modelMatrix[15]);
*/

    /* pre-multiply the light position, as per the orange book, page 216,
     "OpenGL specifies that light positions are transformed by the modelview
     matrix when they are provided to OpenGL..." */
    /* DirectionalLight?  PointLight, SpotLight? */

	// assumes pos[3] = 0.0; only use first 3 of these numbers
	transformf(auxt,pos,modelMatrix);
	auxt[3] = 0.0;

/*
ConsoleMessage("LightToEye, after transformf, auxt %4.2f %4.2f %4.2f %4.2f, pos %4.2f %4.2f %4.2f %4.2f",
auxt[0],auxt[1],auxt[2],auxt[3],
pos[0],pos[1],pos[2],pos[3]);
*/

	for(i=0;i<3;i++){
		pos[i] = auxt[i];
	}
	b = modelMatrix;
	a = dir;
	aux[0] = (float) (b[0]*a[0] +b[4]*a[1] +b[8]*a[2] );
	aux[1] = (float) (b[1]*a[0] +b[5]*a[1] +b[9]*a[2] );
	aux[2] = (float) (b[2]*a[0] +b[6]*a[1] +b[10]*a[2]);
	for(i=0;i<3;i++)
		dir[i] = aux[i];

	// just initialize this to 0.0
	dir[3] = 0.0;

}


/* finished rendering thisshape. */
void finishedWithGlobalShader(void) {
    //printf ("finishedWithGlobalShader\n");


    /* get rid of the shader */
    getAppearanceProperties()->currentShaderProperties = NULL;
FW_GL_BINDBUFFER(GL_ARRAY_BUFFER, 0);

FW_GL_BINDBUFFER(GL_ELEMENT_ARRAY_BUFFER, 0);

}


/* should the system need to rebuild the OpenGL system (eg, Android, 
on restore of screen, iPhone?? Blackberry???) we ensure that the system
state is such that new information will get cached */

void resetGlobalShader() {
	ppRenderFuncs p = (ppRenderFuncs)gglobal()->RenderFuncs.prv;

	//ConsoleMessage ("resetGlobalShader called");

	/* no shader currently active */
	p->currentShader = 0;
}

void restoreGlobalShader(){
	ppRenderFuncs p = (ppRenderFuncs)gglobal()->RenderFuncs.prv;
    if (p->currentShader)
		USE_SHADER(p->currentShader);
}
/* choose and turn on a shader for this geometry */

void enableGlobalShader(s_shader_capabilities_t *myShader) {
	ppRenderFuncs p = (ppRenderFuncs)gglobal()->RenderFuncs.prv;

    //ConsoleMessage ("enableGlobalShader, have myShader %d",myShader->myShaderProgram);
    if (myShader == NULL) {
        finishedWithGlobalShader(); 
        return;
    };
    
    
    getAppearanceProperties()->currentShaderProperties = myShader;
    if (myShader->myShaderProgram != p->currentShader) {
		USE_SHADER(myShader->myShaderProgram);
		p->currentShader = myShader->myShaderProgram;
	}
}


/* send in vertices, normals, etc, etc... to either a shader or via older opengl methods */
void sendAttribToGPU(int myType, int dataSize, int dataType, int normalized, int stride, void *pointer, int texID, char *file, int line){

    s_shader_capabilities_t *me = getAppearanceProperties()->currentShaderProperties;

	// checking to see that we really have the data
	if (me==NULL) 
		return;

#ifdef RENDERVERBOSE

ConsoleMessage ("sendAttribToGPU, getAppearanceProperties()->currentShaderProperties %p\n",getAppearanceProperties()->currentShaderProperties);
ConsoleMessage ("myType %d, dataSize %d, dataType %d, stride %d\n",myType,dataSize,dataType,stride);
	if (me != NULL) {
		switch (myType) {
			case FW_NORMAL_POINTER_TYPE:
				ConsoleMessage ("glVertexAttribPointer  Normals %d at %s:%d\n",me->Normals,file,line);
				break;
			case FW_VERTEX_POINTER_TYPE:
				ConsoleMessage ("glVertexAttribPointer  Vertexs %d at %s:%d\n",me->Vertices,file,line);
				break;
			case FW_COLOR_POINTER_TYPE:
				ConsoleMessage ("glVertexAttribPointer  Colours %d at %s:%d\n",me->Colours,file,line);
				break;
			case FW_TEXCOORD_POINTER_TYPE:
				ConsoleMessage ("glVertexAttribPointer  TexCoords %d at %s:%d\n",me->TexCoords,file,line);
				break;

			default : {ConsoleMessage ("sendAttribToGPU, unknown type in shader\n");}
		}
	}
#endif
#undef RENDERVERBOSE

	switch (myType) {
		case FW_NORMAL_POINTER_TYPE:
		if (me->Normals != -1) {
			glEnableVertexAttribArray(me->Normals);
			glVertexAttribPointer(me->Normals, 3, dataType, normalized, stride, pointer);
		}
			break;
		case FW_FOG_POINTER_TYPE:
		if (me->FogCoords != -1) {
			glEnableVertexAttribArray(me->FogCoords);
			glVertexAttribPointer(me->FogCoords, 1, dataType, normalized, stride, pointer);
		}
			break;

		case FW_VERTEX_POINTER_TYPE:
		if (me->Vertices != -1) {
			glEnableVertexAttribArray(me->Vertices);
			glVertexAttribPointer(me->Vertices, dataSize, dataType, normalized, stride, pointer);
		}
			break;
		case FW_COLOR_POINTER_TYPE:
		if (me->Colours != -1) {
			glEnableVertexAttribArray(me->Colours);
			glVertexAttribPointer(me->Colours, dataSize, dataType, normalized, stride, pointer);
		}
			break;
		case FW_TEXCOORD_POINTER_TYPE:
		if (me->TexCoords[texID] != -1) {
			glEnableVertexAttribArray(me->TexCoords[texID]);
			glVertexAttribPointer(me->TexCoords[texID], dataSize, dataType, normalized, stride, pointer);

		}
			break;
		case FW_CINDEX_POINTER_TYPE:
			if (me->Cindex != -1) {
				//PRINT_GL_ERROR_IF_ANY("");

				glEnableVertexAttribArray(me->Cindex);
				//PRINT_GL_ERROR_IF_ANY("");
				//note I in Attrib I Pointer, opengl 3+, prevents conversion of int to float
				//https://registry.khronos.org/OpenGL-Refpages/gl4/html/glVertexAttribPointer.xhtml
				glVertexAttribIPointer(me->Cindex, 1, dataType, stride, pointer);
				//PRINT_GL_ERROR_IF_ANY("");

			}
			break;
		default : {printf ("sendAttribToGPU, unknown type in shader\n");}
	}
}


void sendBindBufferToGPU (GLenum target, GLuint buffer, char *file, int line) {

	
/*
	if (target == GL_ARRAY_BUFFER_BINDING) printf ("glBindBuffer, GL_ARRAY_BUFFER_BINDING %d at %s:%d\n",buffer,file,line);
	else if (target == GL_ARRAY_BUFFER) printf ("glBindBuffer, GL_ARRAY_BUFFER %d at %s:%d\n",buffer,file,line);
	else if (target == GL_ELEMENT_ARRAY_BUFFER) printf ("glBindBuffer, GL_ELEMENT_ARRAY_BUFFER %d at %s:%d\n",buffer,file,line);
	else printf ("glBindBuffer, %d %d at %s:%d\n",target,buffer,file,line);
	
*/

	glBindBuffer(target,buffer);
}


bool setupShader() {
	return true;
}
void sendFogToShader(s_shader_capabilities_t *me);
void sendClipplanesToShader(s_shader_capabilities_t *me);
int setupShaderB() {

	s_shader_capabilities_t *mysp = getAppearanceProperties()->currentShaderProperties;

PRINT_GL_ERROR_IF_ANY("BEGIN setupShader");
	if (mysp == NULL) 
		return FALSE;

	/* if we had a shader compile problem, do not draw */
	if (!(mysp->compiledOK)) {
#ifdef RENDERVERBOSE
		printf ("shader compile error\n");
#endif
		PRINT_GL_ERROR_IF_ANY("EXIT(false) setupShader");
		return false;
	}

#ifdef RENDERVERBOSE
	printf ("setupShader, we have Normals %d Vertices %d Colours %d TexCoords %d \n",
		mysp->Normals,
		mysp->Vertices,
		mysp->Colours,
		mysp->TexCoords);
#endif

	/* send along lighting, material, other visible properties */
	//PRINT_GL_ERROR_IF_ANY("BEFORE sendFogToShader");
	sendFogToShader(mysp);
	//PRINT_GL_ERROR_IF_ANY("AFTER sendFogToShader");

	sendClipplanesToShader(mysp);
	//PRINT_GL_ERROR_IF_ANY("AFTER sendClipplanesToShader");

	sendMaterialsToShader(mysp); //and lights
	//PRINT_GL_ERROR_IF_ANY("AFTER sendMaterialsToShader");

	sendMatriciesToShader(mysp);
	//PRINT_GL_ERROR_IF_ANY("AFTER sendMatriciesToShader");

	return TRUE;
}

// for particlephysics component we want to be able to do:
// send vbos to shader/gpu
// foreach liveparticle
//    send particle-specific age-related color, texcoords, position to gpu
//    drawOnce()
// clearDraw()
void saveArraysForGPU(int mode, int first, int count){
	draw_call_params params;
	ppRenderFuncs p;
	ttglobal tg = gglobal();
	p = (ppRenderFuncs)tg->RenderFuncs.prv;

	params.calltype = 1;
	params.arrays.arrays_mode = mode;
	params.arrays.arrays_count = count;
	params.arrays.arrays_first = first;
	stack_push(draw_call_params,p->draw_call_params_stack,params);
}


void saveElementsForGPU0(int mode, int count, int type, void *indices){
	//we use a vector/stack because IndexedLineSet and LineSet call several times
	// for one polyline vbo
	draw_call_params params;
	ppRenderFuncs p;
	ttglobal tg = gglobal();
	p = (ppRenderFuncs)tg->RenderFuncs.prv;
	
	params.calltype = 2;
	params.elements.elements_count = count;
	params.elements.elements_mode = mode;
	params.elements.elements_indices = indices;
	params.elements.elements_type = type; //can be UNSIGNED_INT or GL_UNSIGNED_SHORT;
	stack_push(draw_call_params,p->draw_call_params_stack,params);
}
//void saveElementsForGPU(int mode, int count, ushort* indices) {
//	saveElementsForGPU0(mode, count, GL_UNSIGNED_SHORT,indices);
//}
void reallyDrawOnce() {
	//particle system will call this
	//H: this might be a bit like glDrawMultiElements - a list of more primitive triangle fans etc that would make up a 3D shape
	int i;
	draw_call_params* params;
	ppRenderFuncs p;
	ttglobal tg = gglobal();
	p = (ppRenderFuncs)tg->RenderFuncs.prv;
	int loc_side = glGetUniformLocation(p->currentShader,"material_side");
	int twosided = getAppearanceProperties()->twosided;
	int nsides = twosided ? 2 : 1;
	if (!twosided) glUniform1i(loc_side, 0);
	for (int k = 0; k < nsides; k++) {
		if(twosided) glUniform1i(loc_side, k+1); 
		for (i = 0; i < vectorSize(p->draw_call_params_stack); i++) {
			params = vector_get_ptr(draw_call_params, p->draw_call_params_stack, i);
			if (params->calltype == 1) {
				// in msvc you can do try catch in flat C, but not recommended in general - use c++
				// but works when testing/debugging if the video driver is throwing c++ exceptions
				// because we're sending it junk, to stop it from vapor-crashing 
				// https://msdn.microsoft.com/en-us/library/1deeycx5.aspx
#define CATCH_GLDRAWARRAYS_THROWS 1
#if defined(CATCH_GLDRAWARRAYS_THROWS) && defined(_MSC_VER) && defined(W_DEBUG)
				__try {
					glDrawArrays(params->arrays.arrays_mode, params->arrays.arrays_first, params->arrays.arrays_count);
				}
				__except (EXCEPTION_EXECUTE_HANDLER) {
					printf("\n ouch from reallyDrawOnce glDrawArrays \n");
					printf("i= %d n= %d", i, vectorSize(p->draw_call_params_stack));
				}
#else
				glDrawArrays(params->arrays.arrays_mode, params->arrays.arrays_first, params->arrays.arrays_count);
#endif
			}
			else if (params->calltype == 2) {
				glDrawElements(params->elements.elements_mode, params->elements.elements_count, params->elements.elements_type, params->elements.elements_indices);
			}
		}
	}
	//p->draw_call_params_stack->n = 0;
}
void clearDraw(){
	// particlesystem will call this
	ppRenderFuncs p;
	ttglobal tg = gglobal();
	p = (ppRenderFuncs)tg->RenderFuncs.prv;
	p->draw_call_params_stack->n = 0;
}
void reallyDraw(){
	//child_Shape will call this
	reallyDrawOnce();
	clearDraw();
}
void sendArraysToGPU (int mode, int first, int count) {
	#ifdef RENDERVERBOSE
	printf ("sendArraysToGPU start\n"); 
	#endif
    

	// when glDrawArrays bombs it's usually some function left an array
	// enabled that's not supposed to be - try disabling something
	//glDisableClientState(GL_VERTEX_ARRAY);
	//glDisableClientState(GL_NORMAL_ARRAY);
	//glDisableClientState(GL_INDEX_ARRAY);
	//glDisableClientState(GL_COLOR_ARRAY);
	//glDisableClientState(GL_SECONDARY_COLOR_ARRAY);
	//glDisableClientState(GL_FOG_COORDINATE_ARRAY);
	//glDisableClientState(GL_TEXTURE_COORD_ARRAY);
	//glDisableClientState(GL_EDGE_FLAG_ARRAY);

	if (setupShader()){
		profile_start("draw_arr");
//		glDrawArrays(mode,first,count);
		saveArraysForGPU(mode,first,count);
		profile_end("draw_arr");
	}
			#ifdef RENDERVERBOSE
	printf ("sendArraysToGPU end\n"); 
	#endif
}



void sendElementsToGPU (int mode, int count, int *indices) {
    #ifdef RENDERVERBOSE
	printf ("sendElementsToGPU start\n"); 
    #endif
    
	if (setupShader()){
		profile_start("draw_el");
//        glDrawElements(mode,count,GL_UNSIGNED_SHORT,indices);
		saveElementsForGPU0(mode,count,GL_UNSIGNED_INT, indices);
		profile_end("draw_el");
	}

	#ifdef RENDERVERBOSE
	printf ("sendElementsToGPU finish\n"); 
	#endif
}


ttrenderstate renderstate()
{
	ppRenderFuncs p = (ppRenderFuncs)gglobal()->RenderFuncs.prv;
	return stack_top_ptr(struct trenderstate,p->renderstate);
}
void push_new_renderstate() {
	ppRenderFuncs p = (ppRenderFuncs)gglobal()->RenderFuncs.prv;
	struct trenderstate ttr;
	memset(&ttr, 0, sizeof(struct trenderstate));
	stack_push(struct trenderstate, p->renderstate,ttr);
}
void pop_renderstate() {
	ppRenderFuncs p = (ppRenderFuncs)gglobal()->RenderFuncs.prv;
	stack_pop(struct trenderstate, p->renderstate);
}
//true statics:
GLint viewport[4] = {-1,-1,2,2};  //pseudo-viewport - doesn't change, used in glu unprojects
/* These two points (r2,r1) define a ray in pick-veiwport window coordinates 
	r2=viewpoint 
	r1=ray from center of pick-viewport in viewport coordinates
	- in setup_pickray(pick=TRUE,,) the projMatrix is modified for the pick-ray-viewport
	- when unprojecting geometry-local xyz to bearing-local/pick-viewport-local, use pseudo-viewport defined above
*/
struct point_XYZ r1 = {.x=0,.y=0,.z=-1}, r2 = {.x=0,.y=0,.z=0}, r3 = {.x=0,.y=1,.z=0}; //r3 y direction in case needed for testing


struct X3D_Anchor *AnchorsAnchor()
{
	ppRenderFuncs p = (ppRenderFuncs)gglobal()->RenderFuncs.prv;
	return p->AnchorsAnchor;
}
void setAnchorsAnchor(struct X3D_Anchor* anchor)
{
	ppRenderFuncs p = (ppRenderFuncs)gglobal()->RenderFuncs.prv;
	p->AnchorsAnchor = anchor;
}


//static struct currayhit rayph;
//struct currayhit rayHit,rayHitHyper;
/* used to test new hits */



//struct X3D_Group *_rootNode=NULL;	/* scene graph root node */
struct X3D_Node *rootNode()
{
	// ConsoleMessage ("rootNode called");
	ppRenderFuncs p = (ppRenderFuncs)gglobal()->RenderFuncs.prv;	
	if (p==NULL) {
		ConsoleMessage ("rootNode, p null");
		return NULL;
	}
	return p->rootNode;
}
void setRootNode(struct X3D_Node *rn)
{
	ppRenderFuncs p = (ppRenderFuncs)gglobal()->RenderFuncs.prv;
	p->rootNode = rn;
}
//struct Vector *libraries(){
//	ppRenderFuncs p = (ppRenderFuncs)gglobal()->RenderFuncs.prv;
//	if(!p->libraries) p->libraries = newVector(void3 *,1)	;
//	return p->libraries;
//}
//void setLibraries(struct Vector *libvector){
//	//might use this in KILL_oldWorld to NULL the library vector?
//	ppRenderFuncs p = (ppRenderFuncs)gglobal()->RenderFuncs.prv;	
//	p->libraries = libvector;
//}
void addLibrary(char *url, struct X3D_Proto *library, void *res){
	void3 *ul = MALLOC(void3 *,sizeof(void3));
	ppRenderFuncs p = (ppRenderFuncs)gglobal()->RenderFuncs.prv;	
	ul->one = (void *)STRDUP(url);
	ul->two = (void *)library;
	ul->three = res;
	vector_pushBack(void3 *,p->libraries,ul);
}
void3 *librarySearch(char *absoluteUniUrlNoPound){
	ppRenderFuncs p = (ppRenderFuncs)gglobal()->RenderFuncs.prv;	
	void3 *ul;
	struct Vector* libs;
	int n, i;
	libs = p->libraries;
	n = vectorSize(libs);
	for(i=0;i<n;i++){
		ul = vector_get(void3 *,libs,i);
		if(ul)
		if(!strcmp(absoluteUniUrlNoPound,ul->one)){
			return ul; //return res
		}
	}
	return NULL;
}

//void *empty_group=0;

/*******************************************************************************/



void prepare_model_view_pickmatrix0(GLDOUBLE *modelMatrix, GLDOUBLE *mvp){
	//prepares a matrix that will transform a point in geometry-local coordintes
	//into eye/pickray/bearing coordinates ie along the pickray 0,0,1
	GLDOUBLE *pickMatrixi;

	pickMatrixi = getPickrayMatrix(1);

	//pickMatrix is inverted in setup_pickray
	matmultiplyAFFINE(mvp,modelMatrix,pickMatrixi);

}
void prepare_model_view_pickmatrix_inverse0(GLDOUBLE *modelMatrix, GLDOUBLE *mvpi){
	//prepares a matrix that will transform a point in eye/pickray/bearing coords ie 0,0,1
	//into geometry-local coordinates, given the modelMatrix to transform
	// eye/pickray -> geometry-local
	GLDOUBLE mvi[16];
	GLDOUBLE *pickMatrix;

	pickMatrix = getPickrayMatrix(0);

	//pickMatrix is not inverted in setup_pickray
	matinverseAFFINE(mvi,modelMatrix);
	matmultiplyAFFINE(mvpi,pickMatrix,mvi);
}


/* rayhit
	For PointingDeviceSensor component work, called from virt->rendray_<Shape> on VF_Sensitive pass
	- tests if this ray-geometry intersection is closer to the viewpoint than the closest one so far
	- if not it means it is occluded, do nothing
	- if so
	-- updates the closest distance to intersection of pick-ray/bearing with scene geometry so far
	How:
	- the calling rendray_<geometry> function already has the pickray/bearing in its geometry-local
		coordinates, and computes the distance from A as rat and passes it in here.
	- this makes sure its on the B side of A (otherwise its behind the pickray/viewpoint)
	- if its the closest intersection of pickray with scene geometry so far:
		1.records the point, in bearing-local coordinates
		a) for non-sensitive geometry: the point is used to occlude picksensors by being closer to the viewpoint/bearing A
		b) for <Drag>Sensor and TouchSensor, if the point succeeds as the closest point 
			at end of VF_Sensitive pass, the point will be transformed from bearing-local to sensor-local
			to generate eventOuts in do_<>Sensor in sensor-local coordinates
		2.snapshots the sensor's modelview matrix for later use

 */
void rayhit(float rat, float cx,float cy,float cz, float nx,float ny,float nz,
	    float tx,float ty, char *descr)  {
	GLDOUBLE modelMatrix[16];
	ppRenderFuncs p;
	ttglobal tg = gglobal();
	p = (ppRenderFuncs)tg->RenderFuncs.prv;

	/* Real rat-testing */
#ifdef RENDERVERBOSE
	//printf("RAY HIT %s! %f (%f %f %f) (%f %f %f)\n\tR: (%f %f %f) (%f %f %f)\n",
	//       descr, rat,cx,cy,cz,nx,ny,nz,
	//       t_r1.x, t_r1.y, t_r1.z,
	//       t_r2.x, t_r2.y, t_r2.z
	//	);
#endif

	if(rat<0 || (rat>tg->RenderFuncs.hitPointDist && tg->RenderFuncs.hitPointDist >= 0)) {
		return;
	}
	FW_GL_GETDOUBLEV(GL_MODELVIEW_MATRIX, modelMatrix); //snapshot of geometry's modelview matrix
	{
		GLDOUBLE mvp[16];
		struct point_XYZ tp; //note viewpoint/avatar Z=1 behind the viewer, to match the glu_unproject method WinZ = -1
		tp.x = cx; tp.y = cy; tp.z = cz;
		prepare_model_view_pickmatrix0(modelMatrix,mvp);
		transform(&tp,&tp,mvp);
		p->hp = tp;
	}
	tg->RenderFuncs.hitPointDist = rat;
	p->rayHit=p->rayph;
#ifdef RENDERVERBOSE 
	printf ("Rayhit, hp.x y z: - %f %f %f hitPointDist %f %s\n",p->hp.x,p->hp.y,p->hp.z, rat, descr);
#endif
}

/* Call this when modelview and projection modified
	keeps bearing/pick-ray transformed into current geometry-local
	for use in virt->rendray_<geometry> calculations, on VF_Sensitive pass
	bearing-local == pick-viewport-local
*/
void upd_ray0(struct point_XYZ *t_r1, struct point_XYZ *t_r2, struct point_XYZ *t_r3) {
	//struct point_XYZ t_r1,t_r2,t_r3;
	GLDOUBLE modelMatrix[16];
	FW_GL_GETDOUBLEV(GL_MODELVIEW_MATRIX, modelMatrix);
/*
{int i; printf ("\n"); 
printf ("upd_ray, pm %p\n",projMatrix);
for (i=0; i<16; i++) printf ("%4.3lf ",modelMatrix[i]); printf ("\n"); 
for (i=0; i<16; i++) printf ("%4.3lf ",projMatrix[i]); printf ("\n"); 
} 
*/

	{
		//feature-AFFINE_GLU_UNPROJECT
		//FLOPs	112 double:	matmultiplyAFFINE 36, matinverseAFFINE 49, 3x transform (affine) 9 =27
		GLDOUBLE  mvpi[16]; //mvp[16],
		struct point_XYZ r11 = {.x=0.0,.y=0.0,.z=1.0}; //note viewpoint/avatar Z=1 behind the viewer, to match the glu_unproject method WinZ = -1
		{
			//PointSensor needs an original camera axis (not modified pickray camera)
			// to use as a plane normal to intersect the pickray/bearing with
			ttglobal tg;
			double mvi[16];
			struct point_XYZ view_cam_axis, local_cam_axis;
			tg = gglobal();
			view_cam_axis.x = 0.0;
			view_cam_axis.y = 0.0;
			view_cam_axis.z = -1.0;
			matinverseAFFINE(mvi,modelMatrix);
			transformAFFINE(&local_cam_axis,&view_cam_axis,mvi);
			tg->RenderFuncs.camera_axis[0] = local_cam_axis.x;
			tg->RenderFuncs.camera_axis[1] = local_cam_axis.y;
			tg->RenderFuncs.camera_axis[2] = local_cam_axis.z;
			 
		}
		prepare_model_view_pickmatrix_inverse0(modelMatrix, mvpi);
		transform(t_r1,&r11,mvpi);
		transform(t_r2,&r2,mvpi);
		transform(t_r3,&r3,mvpi);
		//r2 is A, r1 is B relative to A in pickray [A,B)
		//we prove it here by moving B along the ray, to distance 1.0 from A, and no change to picking
		//if(0){
		//	vecdiff(t_r1,t_r1,t_r2);
		//	vecnormal(t_r1,t_r1);
		//	vecadd(t_r1,t_r1,t_r2);
		//}
		//printf("Upd_ray new: (%f %f %f) (%f %f %f) \n",	t_r1.x,t_r1.y,t_r1.z,t_r2.x,t_r2.y,t_r2.z);
	}
}
void setup_pickray0();
void upd_ray() {
	ppRenderFuncs p;
	ttglobal tg = gglobal();
	p = (ppRenderFuncs)tg->RenderFuncs.prv;

	setup_pickray0();
	upd_ray0(&p->t_r123.p1,&p->t_r123.p2,&p->t_r123.p3);
	/*
	printf("Upd_ray: (%f %f %f)->(%f %f %f) == (%f %f %f)->(%f %f %f)\n",
	r1.x,r1.y,r1.z,r2.x,r2.y,r2.z,
	t_r1.x,t_r1.y,t_r1.z,t_r2.x,t_r2.y,t_r2.z);
	*/

}
void transformMBB(GLDOUBLE *rMBBmin, GLDOUBLE *rMBBmax, GLDOUBLE *matTransform, GLDOUBLE* inMBBmin, GLDOUBLE* inMBBmax);
int pickrayHitsMBB(struct X3D_Node *node){
	//GOAL: on a render_hier(VF_sensitive) (touch sensor) pass, before checking the ray against geometry, check first if the
	//ray goes through the extent / minimum-bounding-box (MBB) of the shape. If not, no need to check the ray 
	//against all the shape's triangles, speeding up the VF_Sensitive pass.
	//FLOPs 156 double: matmultiplyAffine 36, matInversAffine 48,  transformAffine 8 pts x 12= 72
	GLDOUBLE modelMatrix[16];
	int i, isIn;
	//if using new Sept 2014 pickmatrix, we can test the pickray against the shape node's bounding box
	//and if no hit, then no need to run through rendray testing all triangles
	//feature-AFFINE_GLU_UNPROJECT
	//FLOPs	112 double:	matmultiplyAFFINE 36, matinverseAFFINE 49, 3x transform (affine) 9 =27
	GLDOUBLE mvp[16]; //, mvpi[16];
	GLDOUBLE smin[3], smax[3], shapeMBBmin[3], shapeMBBmax[3];
	int retval;
	retval = TRUE;

	//struct point_XYZ r11 = {0.0,0.0,-1.0}; //note viewpoint/avatar Z=1 behind the viewer, to match the glu_unproject method WinZ = -1
	FW_GL_GETDOUBLEV(GL_MODELVIEW_MATRIX, modelMatrix);

	prepare_model_view_pickmatrix0(modelMatrix, mvp);
	/* generate mins and maxes for avatar cylinder in avatar space to represent the avatar collision volume */
	for(i=0;i<3;i++)
	{
		shapeMBBmin[i] = node->_extent[i*2 + 1];
		shapeMBBmax[i] = node->_extent[i*2];
	}
	transformMBB(smin,smax,mvp,shapeMBBmin,shapeMBBmax); //transform shape's MBB into pickray space
	// the pickray is now at 0,0,x
	isIn = TRUE;
	for(i=0;i<2;i++)
		isIn = isIn && (smin[i] <= 0.0 && smax[i] >= 0.0);
	retval = isIn;
	//printf("%d x %f %f y %f %f\n",isIn,smin[0],smax[0],smin[1],smax[1]);
	//retval = 1;

	return retval;
}




/* if a node changes, void the display lists */
/* Courtesy of Jochen Hoenicke */

void update_node(struct X3D_Node *node) {
	int i;
	static int depth = 0;
	depth++;
	if(depth > 50){
		depth--;
		return;
	}
#ifdef VERBOSE
	printf ("update_node for %d %s nparents %d renderflags %x\n",node, stringNodeType(node->_nodeType),node->_nparents, node->_renderFlags); 
	if (node->_nparents == 0) {
		if (node == rootNode) printf ("...no parents, this IS the rootNode\n"); 
		else printf ("...no parents, this IS NOT the rootNode\n");
	}



	for (i = 0; i < node->_nparents; i++) {
		struct X3D_Node *n = X3D_NODE(node->_parents[i]);
		if( n != 0 ) {
			printf ("	parent %u is %s\n",n,stringNodeType(n->_nodeType));
		} else {
			printf ("	parent %d is NULL\n",i);
		}
	}
#endif

	node->_change ++;

	/* parentVector here yet?? */
	if (node->_parentVector == NULL) {
		return;
	}

	for (i = 0; i < vectorSize(node->_parentVector); i++) {
		struct X3D_Node *n = vector_get(struct X3D_Node *, node->_parentVector,i);
		if(n == node) {
			fprintf(stderr, "Error: self-referential node structure! (node:'%s')\n", stringNodeType(node->_nodeType));
			vector_set(struct X3D_Node*, node->_parentVector, i,NULL);
		} else if( n != 0 ) {
			update_node(n);
		}
	}
	depth--;
	return;
}

/*********************************************************************
 *********************************************************************
 *
 * render_node : call the correct virtual functions to render the node
 * depending on what we are doing right now.
 */

//#ifdef RENDERVERBOSE
//static int renderLevel = 0;
//#endif

#define PRINT_NODE(_node, _v)  do {					\
		if (gglobal()->internalc.global_print_opengl_errors && (gglobal()->display._global_gl_err != GL_NO_ERROR)) { \
			printf("Render_node_v %p (%s) PREP: %p REND: %p CH: %p FIN: %p RAY: %p HYP: %p\n",_v, \
			       stringNodeType(_node->_nodeType),	\
			       _v->prep,				\
			       _v->rend,				\
			       _v->children,			\
			       _v->fin,				\
			       _v->rendray,			\
			       gglobal()->RenderFuncs.hypersensitive);			\
			printf("Render_state geom %d light %d sens %d\n", \
			       renderstate()->render_geom,				\
			       renderstate()->render_light,				\
			       renderstate()->render_sensitive);			\
			printf("pchange %d pichange %d \n", _node->_change, _node->_ichange); \
		}							\
	} while (0)

//static int renderLevel = 0;
//#define RENDERVERBOSE


/* poor-man's performance profiler:
   wrap a section of code like this
     profile_start("section1");
	 ...code...
	 profile_end("section1");
	then let the browser loop for 10 seconds 
	and hit period '.' on the keyboard to get a printout
*/

void profile_start(char *name){
	ppRenderFuncs p;
	struct profile_entry *pe;
	int i, ifound = -1;
	ttglobal tg = gglobal();
	p = (ppRenderFuncs)tg->RenderFuncs.prv;

	if (!p->profiling_on) return;
	pe = p->profile_entries;

	for(i=0;i<p->profile_entry_count;i++){
		if(!strcmp(name,pe[i].name)){
			ifound = i;
			break;
		}
	}
	if(ifound == -1){
		pe[p->profile_entry_count].name = name;
		pe[p->profile_entry_count].hits = 0;
		ifound = p->profile_entry_count;
		p->profile_entry_count++;
	}
	pe[ifound].start = Time1970sec();
}
void profile_end(char *name){
	ppRenderFuncs p;
	struct profile_entry *pe;
	int i, ifound = -1;
	ttglobal tg = gglobal();
	p = (ppRenderFuncs)tg->RenderFuncs.prv;

	if (!p->profiling_on) return;
	pe = p->profile_entries;
	for(i=0;i<p->profile_entry_count;i++){
		if(!strcmp(name,pe[i].name)){
			ifound = i;
			break;
		}
	}
	if(ifound > -1){
		pe[ifound].accum += Time1970sec() - pe[ifound].start;
		pe[ifound].hits++;
	}
}
void profile_print_all(){
	//hit '.' in the graphics window to get here
	ppRenderFuncs p;
	struct profile_entry *pe;
	ttglobal tg = gglobal();
	p = (ppRenderFuncs)tg->RenderFuncs.prv;
	if (!p->profiling_on){
		p->profiling_on = 1;
		ConsoleMessage("turning profiling on\n");
	}else{
		int i;
		pe = p->profile_entries;
		ConsoleMessage("frame rate: %9.3f  number of items tracked: %d\n", gglobal()->Mainloop.BrowserFPS,p->profile_entry_count);
		ConsoleMessage("%15s %10s %15s %10s\n", "profile name", "hits", "time(sec)", "% of 1st");
		for (i = 0; i < p->profile_entry_count; i++){
			ConsoleMessage("%15s %10d %15.3f %10.2f\n", pe[i].name, pe[i].hits, pe[i].accum, pe[i].accum / pe[0].accum*100.0);
		}
	}
}


//unsigned int getShaderFlags(){
shaderflagsstruct getShaderFlags(){
	//return top-of-stack global shaderflags
	//unsigned int retval;
	shaderflagsstruct retval;
	ttglobal tg = gglobal();
	ppRenderFuncs p = (ppRenderFuncs)tg->RenderFuncs.prv;
	//retval = stack_top(unsigned int,p->shaderflags_stack);
	retval = stack_top(shaderflagsstruct,p->shaderflags_stack);
	return retval;
}
void pushShaderFlags(shaderflagsstruct flags){ //unsigned int flags){
	//at root level, before render_hier, you would push 0000000
	//and pop after render_hier
	//for prep_LocalFog you would call this to push (and pop in fin_LocalFog)
	//these flags are for non-leaf-node shader influencers
	//localLights, localFog, clipPlane
	//and will be |= with shape->_shaderTableEntry flags in child_shape
	ttglobal tg = gglobal();
	ppRenderFuncs p = (ppRenderFuncs)tg->RenderFuncs.prv;
	//stack_push(unsigned int,p->shaderflags_stack,flags);
	stack_push(shaderflagsstruct,p->shaderflags_stack,flags);

}
void popShaderFlags(){
	//
	ttglobal tg = gglobal();
	ppRenderFuncs p = (ppRenderFuncs)tg->RenderFuncs.prv;
	//stack_pop(unsigned int,p->shaderflags_stack);
	stack_pop(shaderflagsstruct,p->shaderflags_stack);

}
struct X3D_Node* get_executionContext() {
	//return top-of-stack executionContext
	struct X3D_Node* retval = NULL;
	ttglobal tg = gglobal();
	ppRenderFuncs p = (ppRenderFuncs)tg->RenderFuncs.prv;
	if (p->ectx_stack->n)
		retval = stack_top(struct X3D_Node*, p->ectx_stack);
	return retval;
}
void push_executionContext(struct X3D_Node* broto) {
	//push when entering child_broto (child_proto which is also child_scene, or child_inline), pop on exit
	ttglobal tg = gglobal();
	ppRenderFuncs p = (ppRenderFuncs)tg->RenderFuncs.prv;
	stack_push(struct X3D_Node*, p->ectx_stack, broto);

}
void pop_executionContext() {
	//
	ttglobal tg = gglobal();
	ppRenderFuncs p = (ppRenderFuncs)tg->RenderFuncs.prv;
	stack_pop(struct X3D_Node*, p->ectx_stack);

}
struct X3D_Node *getFogParams(){
	//return top-of-stack Fog or LocalFog
	struct X3D_Node *retval = NULL;
	ttglobal tg = gglobal();
	ppRenderFuncs p = (ppRenderFuncs)tg->RenderFuncs.prv;
	if(p->fog_stack->n)
		retval = stack_top(struct X3D_Node*,p->fog_stack);
	return retval;
}
void pushFogParams(struct X3D_Node *fogparams){
	//at root level, before render_hier, any bound Fog node
	//and pop after render_hier
	//for prep_LocalFog you would call this to push (and pop in fin_LocalFog)
	ttglobal tg = gglobal();
	ppRenderFuncs p = (ppRenderFuncs)tg->RenderFuncs.prv;
	stack_push(struct X3D_Node*,p->fog_stack,fogparams);

}
void popFogParams(){
	//
	ttglobal tg = gglobal();
	ppRenderFuncs p = (ppRenderFuncs)tg->RenderFuncs.prv;
	stack_pop(struct X3D_Node*,p->fog_stack);

}
//struct point_XYZ3 {
//	struct point_XYZ p1;
//	struct point_XYZ p2;
//	struct point_XYZ p3;
//};
void push_ray(){
	//upd_ray();
	//struct point_XYZ t_r1,t_r2,t_r3;
	//struct point_XYZ3 r123;
	ttglobal tg = gglobal();
	ppRenderFuncs p = (ppRenderFuncs)tg->RenderFuncs.prv;
	//r123.p1 = tg->RenderFuncs.t_r1;
	//r123.p2 = tg->RenderFuncs.t_r2;
	//r123.p3 = tg->RenderFuncs.t_r3;

	//stack_push(struct point_XYZ3,p->ray_stack,r123);
	stack_push(struct point_XYZ3,p->ray_stack,p->t_r123);

	//upd_ray0(&t_r1,&t_r2,&t_r3);
	//VECCOPY(tg->RenderFuncs.t_r1,t_r1);
	//VECCOPY(tg->RenderFuncs.t_r2,t_r2);
	//VECCOPY(tg->RenderFuncs.t_r3,t_r3);
	upd_ray0(&p->t_r123.p1,&p->t_r123.p2,&p->t_r123.p3);

}
void pop_ray(){
	//struct point_XYZ t_r1,t_r2,t_r3;
	//struct point_XYZ3 r123;
	ttglobal tg = gglobal();
	ppRenderFuncs p = (ppRenderFuncs)tg->RenderFuncs.prv;
	//upd_ray();
	//r123 = stack_top(struct point_XYZ3,p->ray_stack);
	stack_pop(struct point_XYZ3,p->ray_stack);
	p->t_r123 = stack_top(struct point_XYZ3,p->ray_stack);
	//stack_pop(struct point_XYZ3,p->ray_stack);
	//tg->RenderFuncs.t_r1 = r123.p1;
	//tg->RenderFuncs.t_r2 = r123.p2;
	//tg->RenderFuncs.t_r3 = r123.p3;

}
void get_current_ray(struct point_XYZ* p1, struct point_XYZ* p2){
	ttglobal tg = gglobal();
	ppRenderFuncs p = (ppRenderFuncs)tg->RenderFuncs.prv;
	*p1 = p->t_r123.p1;
	*p2 = p->t_r123.p2;
}
void push_render_geom(int igeom){
	ttglobal tg = gglobal();
	ppRenderFuncs p = (ppRenderFuncs)tg->RenderFuncs.prv;
	ttrenderstate rs = renderstate();
	stack_push(int,p->render_geom_stack,rs->render_geom);
	rs->render_geom = igeom;
}
void pop_render_geom(){
	int igeom;
	ttglobal tg = gglobal();
	ppRenderFuncs p = (ppRenderFuncs)tg->RenderFuncs.prv;
	igeom = stack_top(int,p->render_geom_stack);
	stack_pop(int,p->render_geom_stack);
	ttrenderstate rs = renderstate();
	rs->render_geom = igeom;
}
void push_sensor(struct X3D_Node *node){
	ttglobal tg = gglobal();
	ppRenderFuncs p = (ppRenderFuncs)tg->RenderFuncs.prv;

	push_render_geom(1);
	stack_push(struct currayhit,p->sensor_stack,p->rayph);
	//srh = MALLOC(struct currayhit *,sizeof(struct currayhit));
	////srh = p->rayph;
	//memcpy(srh,&p->rayph,sizeof(struct currayhit));
	p->rayph.hitNode = node; //will be the parent Transform or Group to a PointingDevice (Touch,Drag) Sensor node
	FW_GL_GETDOUBLEV(GL_MODELVIEW_MATRIX, p->rayph.modelMatrix); //snapshot of sensor's modelview matrix
	FW_GL_GETDOUBLEV(GL_PROJECTION_MATRIX, p->rayph.projMatrix);
	//PRINT_GL_ERROR_IF_ANY("render_sensitive"); PRINT_NODE(node,virt);
}
void pop_sensor(){
	ttglobal tg = gglobal();
	ppRenderFuncs p = (ppRenderFuncs)tg->RenderFuncs.prv;

	//memcpy(&p->rayph,srh,sizeof(struct currayhit));
	//FREE_IF_NZ(srh);
	p->rayph = stack_top(struct currayhit,p->sensor_stack);
	stack_pop(struct currayhit,p->sensor_stack);
	pop_render_geom();

}
int getWindex();
int render_foundSelectedViewpoint();
void extent6f_draw(float *extent);
static int draw_extents = TRUE;
int is_vp_new_way();
void wrap_Shape(struct X3D_Node* node);
void* peek_shape();
void render_node(struct X3D_Node *node) {
	struct X3D_Virt *virt;

	//int srg = 0;
	//int sch = 0;
	int justGeom = 0;
	int pushed_ray;
	int pushed_sensor;
	//struct currayhit *srh = NULL;
	ppRenderFuncs p;
	ttglobal tg = gglobal();
	p = (ppRenderFuncs)tg->RenderFuncs.prv;

	X3D_NODE_CHECK(node);
//#define RENDERVERBOSE 1
#ifdef RENDERVERBOSE
	p->renderLevel ++;
#endif

	if(!node) {
#ifdef RENDERVERBOSE
		DEBUG_RENDER("%d no node, quick return\n", renderLevel);
		p->renderLevel--;
#endif
		return;
	}
	virt = virtTable[node->_nodeType];

#ifdef RENDERVERBOSE 
	//printf("%d =========================================NODE RENDERED===================================================\n",renderLevel);
	{
		int i;
		for(i=0;i<p->renderLevel;i++) printf(" ");
	}
	printf("%d node %u (%s) , v %u renderFlags %x ",p->renderLevel, node,stringNodeType(node->_nodeType),virt,node->_renderFlags);

	if ((node->_renderFlags & VF_Viewpoint) == VF_Viewpoint) printf (" VF_Viewpoint");
	if ((node->_renderFlags & VF_Geom )== VF_Geom) printf (" VF_Geom");
	if ((node->_renderFlags & VF_localLight )== VF_localLight) printf (" VF_localLight");
	if ((node->_renderFlags & VF_Sensitive) == VF_Sensitive) printf (" VF_Sensitive");
	if ((node->_renderFlags & VF_Blend) == VF_Blend) printf (" VF_Blend");
	if ((node->_renderFlags & VF_Proximity) == VF_Proximity) printf (" VF_Proximity");
	if ((node->_renderFlags & VF_Collision) == VF_Collision) printf (" VF_Collision");
	if ((node->_renderFlags & VF_globalLight) == VF_globalLight) printf (" VF_globalLight");
	if ((node->_renderFlags & VF_hasVisibleChildren) == VF_hasVisibleChildren) printf (" VF_hasVisibleChildren");
	if ((node->_renderFlags & VF_shouldSortChildren) == VF_shouldSortChildren) printf (" VF_shouldSortChildren");
	if ((node->_renderFlags & VF_Other) == VF_Other) printf (" VF_Other");
	/*
	if ((node->_renderFlags & VF_inPickableGroup == VF_inPickableGroup) printf (" VF_inPickableGroup");
	if ((node->_renderFlags & VF_PickingSensor == VF_PickingSensor) printf (" VF_PickingSensor");
	*/
	printf ("\n");

	//printf("PREP: %d REND: %d CH: %d FIN: %d RAY: %d HYP: %d\n",virt->prep, virt->rend, virt->children, virt->fin,
	//       virt->rendray, hypersensitive);
	//printf("%d state: vp %d geom %d light %d sens %d blend %d prox %d col %d ", renderLevel, 
 //        	render_vp,render_geom,render_light,render_sensitive,render_blend,render_proximity,render_collision); 
	//printf("change %d ichange %d \n",node->_change, node->_ichange);
#endif



	// leaf-node filtering (we still do the transform-children stack)
	// if we are doing Viewpoints, and we don't have a Viewpoint, don't bother doing anything here *
	//if (renderstate()->render_vp == VF_Viewpoint) { 
	if (renderstate()->render_vp == VF_Viewpoint) {
		//if(tg->Bindable.activeLayer == 0)  //no Layerset nodes
		//if ((node->_renderFlags & VF_Viewpoint) != VF_Viewpoint && virt->children == NULL) { 
		if (!is_vp_new_way()) {
			//mystery renderflags not propagated up chain with new way
			// and we're planning do breadth searches when boundvp != selectedvp (including unreachable LOD children)
			// so we'll try without this
			if ((node->_renderFlags & VF_Viewpoint) != VF_Viewpoint) {
#ifdef RENDERVERBOSE
				printf("doing Viewpoint, but this  node is not for us - just returning\n");
				p->renderLevel--;
#endif
				return;
			}
		}
		if(renderstate()->render_vp == VF_Viewpoint && render_foundSelectedViewpoint()){ 
			//on vp pass, just find first DEF/USE of bound viewpoint
			return;
		}
	}

	/* are we working through global PointLights, DirectionalLights or SpotLights, but none exist from here on down? */
	if (renderstate()->render_light) {
		if((node->_renderFlags & VF_globalLight) != VF_globalLight) { 
	#ifdef RENDERVERBOSE
			printf ("doing globalLight, but this  node is not for us - just returning\n"); 
			p->renderLevel--;
	#endif
			return; 
		}
	}
	justGeom = renderstate()->render_geom && !renderstate()->render_sensitive && !renderstate()->render_blend;
	pushed_ray = FALSE;
	pushed_sensor = FALSE;

	if(virt->prep) {
		//transform types will pushmatrix and multiply in their translation.rotation,scale here (and popmatrix in virt->fin)
		DEBUG_RENDER("rs 2\n");
		PRINT_GL_ERROR_IF_ANY("prep start"); PRINT_NODE(node, virt);
		profile_start("prep");
		if(justGeom)
			profile_start("prepgeom");
		virt->prep(node);  
		profile_end("prep");
		if(justGeom)
			profile_end("prepgeom");
		//if(renderstate()->render_sensitive && !tg->RenderFuncs.hypersensitive) {
		//	push_ray(); //upd_ray(); 
		//	pushed_ray = TRUE;
		//}
		PRINT_GL_ERROR_IF_ANY("prep end"); PRINT_NODE(node,virt);
	}
	if(renderstate()->render_sensitive && !tg->RenderFuncs.hypersensitive) {
		push_ray(); //upd_ray(); 
		pushed_ray = TRUE;
	}
	if(renderstate()->render_proximity && virt->proximity) {
		DEBUG_RENDER("rs 2a\n");
		profile_start("proximity");
		virt->proximity(node);
		profile_end("proximity");
		PRINT_GL_ERROR_IF_ANY("render_proximity"); PRINT_NODE(node,virt);
	}
	if(renderstate()->render_geom && ((node->_renderFlags & VF_USE) == VF_USE) && !renderstate()->render_picking){
		//picking sensor, transform sensor and generally any USE_NODE-USE_NODE scenario
		//ideally we would come in here once per scenegraph USE per frame, even when stereo or quad views
		//because we want to work in world coordinates (not view coordinates) so by the time
		//we strip off the view matrix we would have duplicate entries with stereo
		if(getWindex() == 0){
			//just on first window of stereo or quad
			//we don't do this in VF_Proximity because that pass doesn't go all the way to all geom nodes
			//we could give it its own pass, but doing just the first window is a simple hack.
			double modelviewMatrix[16];
			//IF VIEW == 0
			//GL_GET_MODELVIEWMATRIX
			FW_GL_GETDOUBLEV(GL_MODELVIEW_MATRIX, modelviewMatrix);
			//strip viewmatrix - will happen when we invert one of the USEUSE pair, and multiply
			usehit_add2(node,modelviewMatrix,getpickablegroupdata());
		}
	}
	if(renderstate()->render_picking && node->_nodeType == NODE_Shape ){
		//this is for when called from Component_Picking.c on a partial scenegraph,
		//to get geometry nodes in the usehitB list
		//I put vrit->rendray as a way to detect if its geometry, is there a better way?
		double modelviewMatrix[16];
		struct X3D_Shape *shapenode = (struct X3D_Shape*)node;
		if(shapenode->geometry){
			//IF VIEW == 0
			//GL_GET_MODELVIEWMATRIX
			FW_GL_GETDOUBLEV(GL_MODELVIEW_MATRIX, modelviewMatrix);
			//strip viewmatrix - will happen when we invert one of the USEUSE pair, and multiply
			usehitB_add2(shapenode->geometry,modelviewMatrix,getpickablegroupdata());
		}
	}
	
	if(renderstate()->render_collision && virt->collision) {
		DEBUG_RENDER("rs 2b\n");
		profile_start("collision");
		virt->collision(node);
		profile_end("collision");
		PRINT_GL_ERROR_IF_ANY("render_collision"); PRINT_NODE(node,virt);
	}

	if(renderstate()->render_geom && !renderstate()->render_sensitive && !renderstate()->render_picking && virt->rend) {
			DEBUG_RENDER("rs 3\n");
			PRINT_GL_ERROR_IF_ANY("BEFORE render_geom"); PRINT_NODE(node,virt);
			profile_start("rend");
			if (getSAI_X3DNodeType(node->_nodeType) == X3DGeometryNode && peek_shape() == NULL)
				wrap_Shape(node); // printf("geom node\n");
			//void wrap_Shape(struct X3D_Node* node)
			else
				virt->rend(node);
			profile_end("rend");
			PRINT_GL_ERROR_IF_ANY("render_geom"); PRINT_NODE(node,virt);
	}
	if(renderstate()->render_other && virt->other )
	{
		virt->other(node);
	} //other

	if(renderstate()->render_sensitive && ((node->_renderFlags & VF_Sensitive)|| Viewer()->LookatMode ==2)) {
		DEBUG_RENDER("rs 5\n");
		profile_start("sensitive");
		push_sensor(node);
		pushed_sensor = TRUE;
		profile_end("sensitive");
	}
	if(renderstate()->render_geom && renderstate()->render_sensitive && !tg->RenderFuncs.hypersensitive && virt->rendray) {
		DEBUG_RENDER("rs 6\n");
		profile_start("rendray");
		if(pickrayHitsMBB(node))
			virt->rendray(node);
		profile_end("rendray");
		PRINT_GL_ERROR_IF_ANY("rs 6"); PRINT_NODE(node,virt);
	}

	/* May 16 2016: now we don't come into render_hier on hypersensitive
    if((renderstate()->render_sensitive) && (tg->RenderFuncs.hypersensitive == node)) {
		DEBUG_RENDER("rs 7\n");
		p->hyper_r1 = p->t_r123.p1; //tg->RenderFuncs.t_r1;
		p->hyper_r2 = p->t_r123.p2; //tg->RenderFuncs.t_r2;
		tg->RenderFuncs.hyperhit = 1;
    }
	*/

	/* start recursive section */
    if(virt->children) { 
		DEBUG_RENDER("rs 8 - has valid child node pointer\n");
		//if(! (renderstate()->render_vp == VF_Viewpoint && render_foundLayerViewpoint())){ //on vp pass, just find first DEF/USE of bound viewpoint
			//printf("children ");
			virt->children(node);
		//}
		//}else{
		//	printf("skipping ");
		//}
		PRINT_GL_ERROR_IF_ANY("children"); PRINT_NODE(node,virt);
    }
	/* end recursive section */

	if(renderstate()->render_other && virt->other)
	{
	}

	if(pushed_sensor)
		pop_sensor();

	if(virt->fin) {
		DEBUG_RENDER("rs A\n");
		profile_start("fin");
		if(justGeom)
			profile_start("fingeom");

		virt->fin(node);
		profile_end("fin");
		if(justGeom)
			profile_end("fingeom");
		//if(renderstate()->render_sensitive && virt == &virt_Transform) {
		//	upd_ray();
		//}
		PRINT_GL_ERROR_IF_ANY("fin"); PRINT_NODE(node,virt);
	}
	if(pushed_ray)
		pop_ray();

#ifdef RENDERVERBOSE 
	{
		int i;
		for(i=0;i<p->renderLevel;i++)printf(" ");
	}
	printf("%d (end render_node)\n",p->renderLevel);
	p->renderLevel--;
#endif
}
//#undef RENDERVERBOSE

/*
 * The following code handles keeping track of the parents of a given
 * node. This enables us to traverse the scene on C level for optimizations.
 *
 * We use this array code because VRML nodes usually don't have
 * hundreds of children and don't usually shuffle them too much.
 */
//dug9 dec 13 >>
struct X3D_Node* getTypeNode(struct X3D_Node *node);

void add_parent(struct X3D_Node *node, struct X3D_Node *parent, char *file, int line) {
	struct X3D_Node* itype;
	if(!node) return;
	//if(node->_nodeType == NODE_PlaneSensor)
	//	printf("hi from add_parent, have a Planesensor");
#ifdef CHILDVERBOSE
	printf ("add_parent; adding node %u ,to parent %u at %s:%d\n",node,  parent,file,line);
	printf ("add_parent; adding node %x ,to parent %x (hex) at %s:%d\n",node,  parent,file,line);
	printf ("add_parent; adding node %p ,to parent %p (ptr) at %s:%d\n",node,  parent,file,line);


	printf ("add_parent; adding node %u (%s) to parent %u (%s) at %s:%d\n",node, stringNodeType(node->_nodeType), 
		parent, stringNodeType(parent->_nodeType),file,line);
#endif

	parent->_renderFlags = parent->_renderFlags | node->_renderFlags;

	/* add it to the parents list */
	for(int i=0;i<node->_parentVector->n;i++)
	{
		struct X3D_Node * parent2 = vector_get(struct X3D_Node*,node->_parentVector,i);
		if(parent == parent2){
			//printf("ouch adding the same parent twice\n");
			return;
		}
	}
	vector_pushBack (struct X3D_Node*,node->_parentVector, parent);
	/* tie in sensitive nodes */
	itype = getTypeNode(node);
	if(itype)
		setSensitive (parent, itype );
}

void remove_parent(struct X3D_Node *child, struct X3D_Node *parent) {
	int i;
	int pi;

	if(!child) return;
	if(!parent) return;
	
	//JAS printf ("remove_parent, parent %p (%s) , child %p (%s)\n",parent, stringNodeType(parent->_nodeType),
	//JAS 	child, stringNodeType(child->_nodeType));

	//JAS printf ("remove_parent, parent vector size %d\n",vectorSize(child->_parentVector));

	pi = -1;
	for (i=0; i<vectorSize(child->_parentVector); i++) {
		struct X3D_Node *n = vector_get(struct X3D_Node *, child->_parentVector,i);
		//JAS printf ("remPar, ele %d was %p\n",i,n);
		if (n==parent) pi = i;
	}

	if (pi >=0) {
		//JAS printf ("remove parent, pi %d\n",pi);

		struct X3D_Node *n = vector_get(struct X3D_Node *, child->_parentVector,vectorSize(child->_parentVector)-1);

		/* get the last entry, and overwrite the entry found */
		vector_set(struct X3D_Node*, child->_parentVector, pi,n);

		/* take that last entry off the vector */
		vector_popBack(struct X3D_Node*, child->_parentVector);
	}

	//JAS - verification that parent was indeed removed.
	//JAS printf ("remove_parent, after parent vector size %d\n",vectorSize(child->_parentVector));
	//JAS for (i=0; i<vectorSize(child->_parentVector); i++) {
	//JAS 	struct X3D_Node *n = vector_get(struct X3D_Node *, child->_parentVector,i);
	//JAS 	printf ("remPar, ele %d was %p\n",i,n);
	//JAS }
}


#include "../x3d_parser/Bindable.h"
int fwl_getShadingStyle();
void push_globalRenderFlags(){
	//call in render_hier for geom or blend passes
	//unsigned int shaderflags;
	shaderflagsstruct shaderflags;
	ttglobal tg = gglobal();
	shaderflags = getShaderFlags(); //take a copy (which should be 0000000 at root level)
	memset(&shaderflags,0,sizeof(shaderflagsstruct));
	//modify copy
	//A. if there's a bound (non local) fog, copy its state to fog_state
	if(vectorSize(getActiveBindableStacks(tg)->fog) > 0){
		//there's a bound fog, bound fogs are enabled
		struct X3D_Fog *fog = stack_top(struct X3D_Fog*,getActiveBindableStacks(tg)->fog);
		if(fog->visibilityRange > 0.0f){
			//enabled
			//set fog bit in renderflags
			shaderflags.base |= FOG_APPEARANCE_SHADER;
			//push fogparams
			pushFogParams((struct X3D_Node*)fog);
		}
	}
	//B. Anaglyph?
	if(Viewer()->anaglyph || Viewer()->anaglyphB)
		shaderflags.base |= WANT_ANAGLYPH;

	//C. ShadingStyle ie 0 flat, 1 gouraud, 2 phong, 3 wire
	switch(fwl_getShadingStyle()){
		case 0: shaderflags.base |= SHADINGSTYLE_FLAT; break;
		case 1: shaderflags.base |= SHADINGSTYLE_GOURAUD; break;
		case 2: shaderflags.base |= SHADINGSTYLE_PHONG; break;
		case 3: shaderflags.base |= SHADINGSTYLE_WIRE; break;
		default:
			shaderflags.base |= SHADINGSTYLE_PHONG; break;
	}
	if(tg->Component_TextureProjector.globalProjector){
		shaderflags.base |= HAVE_PROJECTIVETEXTURE;
	}
	pushShaderFlags(shaderflags); //push nodified copy
}
void pop_globalRenderFlags(){
	//call after render_hier for geom or blend passes
	ttglobal tg = gglobal();

	popShaderFlags(); //pop modified copy

	//A. Fog, pop its stack only if it was pushed in push_globalRenderFlags
	if(vectorSize(getActiveBindableStacks(tg)->fog) > 0){
		struct X3D_Fog *fog = stack_top(struct X3D_Fog*,getActiveBindableStacks(tg)->fog);
		if(fog->visibilityRange > 0.0f){
			//enabled
			//pop fogParms
			popFogParams();
		}
	}
	tg->Component_TextureProjector.globalProjector = 0; //watch outL if you do ashort-cut stereo with 2 render_heir(geom) then this shoulod be zeroed after last one or on next frame start

}
struct what_string {
int iwhat;
char *cwhat;
} what_strings [] = {
{VF_Viewpoint,"Viewpoint"},
{VF_Geom,"Geom"},
{VF_globalLight,"globalLight"},
{VF_Sensitive,"Sensitive"},
{VF_Picking,"Picking"},
{VF_Blend,"Blend"},
{VF_Proximity,"Proximit"},
{VF_Collision,"Collision"},
{VF_Other,"Other"},
{VF_Cube,"Cube"},
{VF_Background,"Background"},
{0,NULL},
};
void rwhat_printf(int rwhat){
	struct what_string *ws;
	int k = 0;
	ws = &what_strings[k];
	while(ws->cwhat){
		if(rwhat & ws->iwhat) printf("%s ",ws->cwhat);
		k++;
		ws = &what_strings[k];
	}
	if (k) printf("\n");

}
void render_headlight();
void clear_vp_reachable_flags();
void render_hier(struct X3D_Node *g, int rwhat) {
	/// not needed now - see below struct point_XYZ upvec = {0,1,0};
	/// not needed now - see below GLDOUBLE modelMatrix[16];

	ppRenderFuncs p;
	shaderflagsstruct shaderflags;
	ttglobal tg = gglobal();
	ttrenderstate rs;
	p = (ppRenderFuncs)tg->RenderFuncs.prv;
	rs = renderstate();
	memset(&shaderflags,0,sizeof(shaderflagsstruct));
	pushShaderFlags(shaderflags);

	
	/*
	printf ("start of render_hier, rwhat %x, node has %x ",rwhat, g->_renderFlags);
	if ((g->_renderFlags & VF_Viewpoint) == VF_Viewpoint) printf ("VF_Viewpoint ");
	if ((g->_renderFlags & VF_Geom) == VF_Geom) printf ("VF_Geom ");
	if ((g->_renderFlags & VF_localLight) == VF_localLight) printf ("VF_localLight ");
	if ((g->_renderFlags & VF_Sensitive) == VF_Sensitive) printf ("VF_Sensitive ");
	if ((g->_renderFlags & VF_Blend) == VF_Blend) printf ("VF_Blend ");
	if ((g->_renderFlags & VF_Proximity) == VF_Proximity) printf ("VF_Proximity ");
	if ((g->_renderFlags & VF_Collision) == VF_Collision) printf ("VF_Collision ");
	if ((g->_renderFlags & VF_globalLight) == VF_globalLight) printf ("VF_globalLight ");
	if ((g->_renderFlags & VF_hasVisibleChildren) == VF_hasVisibleChildren) printf ("VF_hasVisibleChildren ");
	printf ("\n");
	*/

	rs->render_vp = rwhat & VF_Viewpoint;
	rs->render_geom =  rwhat & VF_Geom;
	rs->render_light = rwhat & VF_globalLight;
	rs->render_sensitive = rwhat & VF_Sensitive;
	rs->render_picking = rwhat & VF_Picking;
	rs->render_blend = rwhat & VF_Blend;
	rs->render_proximity = rwhat & VF_Proximity;
	rs->render_collision = rwhat & VF_Collision;
	rs->render_other = rwhat & VF_Other;
	rs->render_cube = rwhat & VF_Cube;
	rs->render_background = rwhat & VF_Background;
	rs->render_depth = rwhat & VF_Depth;
	rs->rwhat = rwhat;

	//printf ("render_hier, render_geom %x render_blend %x\n",rs->render_geom, rs->render_blend);


	tg->RenderFuncs.hitPointDist = -1;


#ifdef RENDERVERBOSE
	 printf ("render_hier vp %d geom %d light %d sens %d blend %d prox %d col %d\n",
	   rs->render_vp,rs->render_geom,rs->render_light,rs->render_sensitive,rs->render_blend,rs->render_proximity,rs->render_collision);  
#endif

	if (!g) {
		/* we have no geometry yet, sleep for a tiny bit */
		usleep(1000);
		return;
	}

#ifdef RENDERVERBOSE
	printf("Render_hier node=%d what=%d\n", g, rwhat);
#endif

	if (rs->render_light) {
		render_headlight();
	}
	if (rs->render_sensitive) {
		upd_ray();
	}
	if(rs->render_blend || rs->render_geom){
		push_globalRenderFlags();
	}
	profile_start("render_hier");
	//push_group_extent_default();
	render_node(X3D_NODE(g));
	//pop_group_extent(); // up where parents are
	//rwhat_printf(rwhat);
	profile_end("render_hier");
	if(rs->render_blend || rs->render_geom){
		pop_globalRenderFlags();
	}
	popShaderFlags();


}
void render_hier2(struct X3D_Node* g, int rwhat) {
	// used for sub-scenegraph rendering ie depth maps from render_Light > render_shadowmap

	ppRenderFuncs p;
	shaderflagsstruct shaderflags;
	ttglobal tg = gglobal();
	ttrenderstate rs;
	p = (ppRenderFuncs)tg->RenderFuncs.prv;
	push_new_renderstate();
	rs = renderstate();
	memset(&shaderflags, 0, sizeof(shaderflagsstruct));
	pushShaderFlags(shaderflags);


	/*
	printf ("start of render_hier, rwhat %x, node has %x ",rwhat, g->_renderFlags);
	if ((g->_renderFlags & VF_Viewpoint) == VF_Viewpoint) printf ("VF_Viewpoint ");
	if ((g->_renderFlags & VF_Geom) == VF_Geom) printf ("VF_Geom ");
	if ((g->_renderFlags & VF_localLight) == VF_localLight) printf ("VF_localLight ");
	if ((g->_renderFlags & VF_Sensitive) == VF_Sensitive) printf ("VF_Sensitive ");
	if ((g->_renderFlags & VF_Blend) == VF_Blend) printf ("VF_Blend ");
	if ((g->_renderFlags & VF_Proximity) == VF_Proximity) printf ("VF_Proximity ");
	if ((g->_renderFlags & VF_Collision) == VF_Collision) printf ("VF_Collision ");
	if ((g->_renderFlags & VF_globalLight) == VF_globalLight) printf ("VF_globalLight ");
	if ((g->_renderFlags & VF_hasVisibleChildren) == VF_hasVisibleChildren) printf ("VF_hasVisibleChildren ");
	printf ("\n");
	*/

	rs->render_vp = rwhat & VF_Viewpoint;
	rs->render_geom = rwhat & VF_Geom;
	rs->render_light = rwhat & VF_globalLight;
	rs->render_sensitive = rwhat & VF_Sensitive;
	rs->render_picking = rwhat & VF_Picking;
	rs->render_blend = rwhat & VF_Blend;
	rs->render_proximity = rwhat & VF_Proximity;
	rs->render_collision = rwhat & VF_Collision;
	rs->render_other = rwhat & VF_Other;
	rs->render_cube = rwhat & VF_Cube;
	rs->render_background = rwhat & VF_Background;
	rs->render_depth = rwhat & VF_Depth;

	//printf ("render_hier, render_geom %x render_blend %x\n",rs->render_geom, rs->render_blend);


	//tg->RenderFuncs.hitPointDist = -1;


#ifdef RENDERVERBOSE
	printf("render_hier vp %d geom %d light %d sens %d blend %d prox %d col %d\n",
		rs->render_vp, rs->render_geom, rs->render_light, rs->render_sensitive, rs->render_blend, rs->render_proximity, rs->render_collision);
#endif

	if (!g) {
		/* we have no geometry yet, sleep for a tiny bit */
		//usleep(1000);
		return;
	}

#ifdef RENDERVERBOSE
	printf("Render_hier node=%d what=%d\n", g, rwhat);
#endif

	if (rs->render_light) {
		render_headlight();
	}
	//if (rs->render_sensitive) {
	//	upd_ray();
	//}
	//if (rs->render_blend || rs->render_geom) {
	//	push_globalRenderFlags();
	//}
	//if (rs->render_geom)
	//	clear_vp_reachable_flags();
	//profile_start("render_hier");
	//push_group_extent_default();
	render_node(X3D_NODE(g));
	//pop_group_extent(); // up where parents are
	//rwhat_printf(rwhat);
	//profile_end("render_hier");
	//if (rs->render_blend || rs->render_geom) {
	//	pop_globalRenderFlags();
	//}
	popShaderFlags();
	pop_renderstate();

}
void clear_renderstate(){
	ppRenderFuncs p;
	ttglobal tg = gglobal();
	ttrenderstate rs;
	p = (ppRenderFuncs)tg->RenderFuncs.prv;
	rs = renderstate();
	memset(rs,0,sizeof(ttrenderstate));
}

/******************************************************************************
 *
 * shape compiler "thread"
 *
 ******************************************************************************/

void compileNode (void (*nodefn)(void *, void *, void *, void *, void *, void *), void *node, void *Icoord, void *IfogCoord, void *Icolor, void *Inormal, void *ItexCoord) {
	void *coord; void *fogCoord; void *color; void *normal; void *texCoord;

	/* are any of these SFNodes PROTOS? If so, get the underlying real node, as PROTOS are handled like Groups. */
	POSSIBLE_PROTO_EXPANSION(void *, Icoord,coord)
	POSSIBLE_PROTO_EXPANSION(void *, IfogCoord,fogCoord)
	POSSIBLE_PROTO_EXPANSION(void *, Icolor,color)
	POSSIBLE_PROTO_EXPANSION(void *, Inormal,normal)
	POSSIBLE_PROTO_EXPANSION(void *, ItexCoord,texCoord)

	nodefn(node, coord, fogCoord, color, normal, texCoord);
}

void do_NurbsPositionInterpolator (void *node);
void do_NurbsOrientationInterpolator (void *node);
void do_NurbsSurfaceInterpolator (void *node);
/* for CRoutes, we need to have a function pointer to an interpolator to run, if we
   route TO an interpolator */
void *returnInterpolatorPointer (int nodeType) {
	void (*do_interp)(void *);

	do_interp = NULL;
	switch(nodeType){
		case NODE_OrientationInterpolator: do_interp = do_Oint4; break;
		case NODE_CoordinateInterpolator2D: do_interp = do_OintCoord2D; break;
		case NODE_PositionInterpolator2D: do_interp = do_OintPos2D; break;
		case NODE_ScalarInterpolator: do_interp = do_OintScalar; break;
		case NODE_ColorInterpolator: do_interp = do_ColorInterpolator; break;
		case NODE_PositionInterpolator: do_interp = do_PositionInterpolator; break;
		case NODE_CoordinateInterpolator: do_interp = do_OintCoord; break;
		case NODE_VectorInterpolator: do_interp = do_OintVector; break;
		case NODE_CoordinateMorpher: do_interp = do_CoordinateMorph; break;
		case NODE_NormalMorpher: do_interp = do_NormalMorph; break;
		case NODE_NormalInterpolator: do_interp = do_OintNormal; break;
		case NODE_EaseInEaseOut: do_interp = do_EaseInEaseOut; break;
		case NODE_SplinePositionInterpolator: do_interp = do_SplinePositionInterpolator; break;
		case NODE_SplinePositionInterpolator2D: do_interp = do_SplinePositionInterpolator2D; break;
		case NODE_SplineScalarInterpolator: do_interp = do_SplineScalarInterpolator; break;
		case NODE_SquadOrientationInterpolator: do_interp =  do_SquadOrientationInterpolator; break;
		case NODE_GeoPositionInterpolator: do_interp = do_GeoPositionInterpolator; break;
		case NODE_NurbsPositionInterpolator: do_interp = do_NurbsPositionInterpolator; break;
		case NODE_NurbsOrientationInterpolator: do_interp = do_NurbsOrientationInterpolator; break;
		case NODE_NurbsSurfaceInterpolator: do_interp = do_NurbsSurfaceInterpolator; break;
		case NODE_BooleanFilter: do_interp = do_BooleanFilter; break;
		case NODE_BooleanSequencer: do_interp = do_BooleanSequencer; break;
		case NODE_BooleanToggle: do_interp = do_BooleanToggle; break;
		case NODE_BooleanTrigger: do_interp = do_BooleanTrigger; break;
		case NODE_IntegerTrigger: do_interp = do_IntegerTrigger; break;
		case NODE_IntegerSequencer: do_interp = do_IntegerSequencer; break;
		case NODE_TimeTrigger: do_interp = do_TimeTrigger; break;
		case NODE_GeoConvert: do_interp = do_GeoConvert; break;
		default:
			do_interp = NULL;
	}
	return (void *)do_interp;
}
/*
void *returnInterpolatorPointer (const char *x) {
	if (strcmp("OrientationInterpolator",x)==0) { return (void *)do_Oint4;
	} else if (strcmp("CoordinateInterpolator2D",x)==0) { return (void *)do_OintCoord2D;
	} else if (strcmp("PositionInterpolator2D",x)==0) { return (void *)do_OintPos2D;
	} else if (strcmp("ScalarInterpolator",x)==0) { return (void *)do_OintScalar;
	} else if (strcmp("ColorInterpolator",x)==0) { return (void *)do_ColorInterpolator;
	} else if (strcmp("PositionInterpolator",x)==0) { return (void *)do_PositionInterpolator;
	} else if (strcmp("CoordinateInterpolator",x)==0) { return (void *)do_OintCoord;
	} else if (strcmp("NormalInterpolator",x)==0) { return (void *)do_OintNormal;
	} else if (strcmp("GeoPositionInterpolator",x)==0) { return (void *)do_GeoPositionInterpolator;
	} else if (strcmp("NurbsPositionInterpolator",x)==0) { return (void *)do_NurbsPositionInterpolator;
	} else if (strcmp("NurbsOrientationInterpolator",x)==0) { return (void *)do_NurbsOrienatationInterpolator;
	} else if (strcmp("NurbsSurfaceInterpolator",x)==0) { return (void *)do_NurbsSurfaceInterpolator;
	} else if (strcmp("BooleanFilter",x)==0) { return (void *)do_BooleanFilter;
	} else if (strcmp("FloatMultiply",x)==0) { return (void *)do_FloatMultiply;
	} else if (strcmp("BooleanSequencer",x)==0) { return (void *)do_BooleanSequencer;
	} else if (strcmp("BooleanToggle",x)==0) { return (void *)do_BooleanToggle;
	} else if (strcmp("BooleanTrigger",x)==0) { return (void *)do_BooleanTrigger;
	} else if (strcmp("IntegerTrigger",x)==0) { return (void *)do_IntegerTrigger;
	} else if (strcmp("IntegerSequencer",x)==0) { return (void *)do_IntegerSequencer;
	} else if (strcmp("TimeTrigger",x)==0) { return (void *)do_TimeTrigger;
	
	} else {
		return 0;
	}
}
*/



void checkParentLink (struct X3D_Node *node,struct X3D_Node *parent) {
        int n;

	int *offsetptr;
	char *memptr;
	struct Multi_Node *mfn;
	uintptr_t *voidptr;

        if (node == NULL) return;

	/* printf ("checkParentLink for node %u parent %u type %s\n",node,parent,stringNodeType(node->_nodeType)); */
 
        if (parent != NULL) ADD_PARENT(node, parent);

	if ((node->_nodeType<0) || (node->_nodeType>NODES_COUNT)) {
		ConsoleMessage ("checkParentLink - %d not a valid nodeType",node->_nodeType);
		return;
	}

	/* find all the fields of this node */
	offsetptr = (int *) NODE_OFFSETS[node->_nodeType];

	/* FIELDNAMES_bboxCenter, offsetof (struct X3D_Group, bboxCenter),  FIELDTYPE_SFVec3f, KW_field, */
	while (*offsetptr >= 0) {

		/* 
		   printf ("	field %s",FIELDNAMES[offsetptr[0]]); 
		   printf ("	offset %d",offsetptr[1]);
		   printf ("	type %s",FIELDTYPES[offsetptr[2]]);
		   printf ("	kind %s\n",KEYWORDS[offsetptr[3]]);
		*/

		/* worry about SFNodes and MFNodes */
		if ((offsetptr[2] == FIELDTYPE_SFNode) || (offsetptr[2] == FIELDTYPE_MFNode)) {
			if ((offsetptr[3] == KW_initializeOnly) || (offsetptr[3] == KW_inputOutput)) {

				/* create a pointer to the actual field */
				memptr = (char *) node;
				memptr += offsetptr[1];

				if (offsetptr[2] == FIELDTYPE_SFNode) {
					/* get the field as a POINTER VALUE, not just a pointer... */
					voidptr = (uintptr_t *) memptr;
					voidptr = (uintptr_t *) *voidptr;

					/* is there a node here? */
					if (voidptr != NULL) {
						checkParentLink(X3D_NODE(voidptr),node);
					}
				} else {
					mfn = (struct Multi_Node*) memptr;
					/* printf ("MFNode has %d children\n",mfn->n); */
					for (n=0; n<mfn->n; n++) {
				                checkParentLink(mfn->p[n],node);
					}
				}
			}

		}
		offsetptr += FIELDOFFSET_LENGTH;
	}
}

#define X3D_COORD(node) ((struct X3D_Coordinate*)node)
#define X3D_GEOCOORD(node) ((struct X3D_GeoCoordinate*)node)

/* get a coordinate array - (SFVec3f) from either a NODE_Coordinate or NODE_GeoCoordinate */
struct Multi_Vec3f *getCoordinate (struct X3D_Node *innode, char *str) {
	struct X3D_Coordinate * xc;
	struct X3D_GeoCoordinate *gxc;
	struct X3D_Node *node;

	POSSIBLE_PROTO_EXPANSION (struct X3D_Node *, innode,node)

	if(node){
		xc = X3D_COORD(node);
		/* printf ("getCoordinate, have a %s\n",stringNodeType(xc->_nodeType)); */
		if (xc->_nodeType == NODE_Coordinate) {
			return &(xc->point);
		} else if (xc->_nodeType == NODE_GeoCoordinate) {
			COMPILE_IF_REQUIRED_RETURN_NULL_ON_ERROR;
			gxc = X3D_GEOCOORD(node);
			return &(gxc->__movedCoords);
		} else {
			ConsoleMessage ("%s - coord expected but got %s\n", stringNodeType(xc->_nodeType));
		}
	}
	return NULL;
}



/*
void printLTDebug(char * fn, int ln)
{
 ppRenderFuncs p = (ppRenderFuncs)gglobal()->RenderFuncs.prv;
ConsoleMessage ("headlight pos %f %f %f %f at %s:%d",
p->light_pos[HEADLIGHT_LIGHT][0],p->light_pos[HEADLIGHT_LIGHT][1],
                p->light_pos[HEADLIGHT_LIGHT][2],p->light_pos[HEADLIGHT_LIGHT][3],fn,ln);
}
*/

